import gzip
import hashlib
import re
from collections import Counter
from itertools import product
from typing import List, Optional, Dict

from time import  sleep
from urllib import parse
from os import path
import requests
import json
from pathlib import Path
import copy
import csv
import networkx as nx
from tqdm import tqdm
from pybiblionet.openalex.utils import check_date_format, abstract_inverted_index_to_abstract


def string_generator_from_lite_regex(queries_regex: str) -> List[str]:
    """
    Generates a list of query strings from a simplified regular expression pattern.

    This function takes a simplified regular expression pattern, extracts groups
    of alternatives, and produces all possible combinations of these alternatives
    as separate query strings. Each resulting query string is enclosed in double
    quotes.

    :param queries_regex: A string representing a simplified regular expression pattern.
                          The pattern should contain groups of alternatives enclosed in
                          parentheses and separated by the pipe character '|'.
                          Example: r"(1|2|3)(a|b|c)"

    :return: A list of strings, each representing a possible combination of the alternatives
             defined in the input pattern. Each string is enclosed in double quotes.

    Example:

    >>> queries_regex = r'"(1|3|5|10|15|20|30|45|one|three|five|ten|fifteen|twenty|thirty|fortyfive|x)( )(minute|minutes|min)( )(city|neighborhood|neighbourhood|urban|town)"'
    >>> result = string_generator_from_lite_regex(queries_regex)
    >>> print(result)
    ['"1 minute city"', '"1 minute neighborhood"', '"1 minute neighbourhood"', ...]

    """
    groups=[]
    for group in re.findall(r"\([ a-zA-Z0-9\-\|]{1,}\)",queries_regex):
        groups.append(group.replace("(","").replace(")","").split("|"))

    queries=[]
    for item in product(*groups):
        if len(item)==len(groups):
            queries.append("\""+''.join(item)+"\"")

    return queries

def query_OpenAlex(api_type: str, parameter: Optional[str], mail: str, api_key: str,
                   from_publication_date: Optional[str] = None,
                   to_publication_date: Optional[str] = None,
                   cache: bool = True) -> List[Dict[str, Dict]]:
    """
    Queries the OpenAlex API for works based on the type of query (search, cite, or cited_by).

    :param api_type: The type of API query ('search', 'cite', or 'cited_by').
    :param parameter: The article ID for 'cite' and 'cited_by' queries. Search string for 'search' queries.
    :param mail: The email address to be included in the API request for identification.
    :param from_publication_date: Optional start date for filtering results (format: 'YYYY-MM-DD').
    :param to_publication_date: Optional end date for filtering results (format: 'YYYY-MM-DD').
    :param cache: Boolean indicating whether to use caching for API responses (default True).
    :return: A list of results matching the query.
    """
    if api_type == 'search':
        cache_path = Path("cache") / "search"
        base_url = f'https://api.openalex.org/works?filter=title_and_abstract.search:{parse.quote_plus(parameter)}'
    elif api_type == 'cite':
        cache_path = Path("cache") / "cite"
        parameter = parameter.split("/")[-1]
        base_url = f'https://api.openalex.org/works?filter=cites:{parameter}'
    elif api_type == 'cited_by':
        cache_path = Path("cache") / "cite_by"
        parameter = parameter.split("/")[-1]
        base_url = f'https://api.openalex.org/works?filter=cited_by:{parameter}'
    else:
        raise ValueError("Invalid api_type. Must be 'search', 'cite', or 'cited_by'.")

    # Ensure the directory exists
    cache_path.mkdir(parents=True, exist_ok=True)

    if from_publication_date and check_date_format(from_publication_date):
        base_url += f',from_publication_date:{from_publication_date}'
    if to_publication_date and check_date_format(to_publication_date):
        base_url += f',to_publication_date:{to_publication_date}'

    base_url += f'&per-page=200&mailto={mail}'
    cursor = "*"
    results = []
    Path(cache_path).mkdir(parents=True, exist_ok=True)

    while cursor:
        try:
            this_url = f'{base_url}&cursor={cursor}'
            cursor = False
            print(this_url)
            def hash_url(url:str)->str:
                return hashlib.md5(url.encode('utf-8')).hexdigest()
            this_url_filename=this_url.split('?')[1]
            filename = cache_path / f"{hash_url(this_url_filename)}.json.gz"
            #filename = cache_path / f"{this_url.split('?')[-1]}.json.gz"

            if cache and path.isfile(filename):
                with gzip.open(filename, 'rt', encoding='utf-8') as f:
                    result = json.load(f)
            else:
                response = requests.get(this_url+f"&api_key={api_key}")
                status_code = response.status_code
                if str(status_code) == '429':
                    print(f"Daily limit exceeded. Wait 300 seconds before trying again.")
                    sleep(300)
                elif str(status_code) == '403':
                    print(f"Rate limit exceeded. Slow down. Wait 60 seconds before trying again.")
                    sleep(60)
                response.raise_for_status()
                txt = response.text

                if cache:
                    with gzip.open(filename, 'wt', encoding='utf-8') as f:
                        f.write(txt)

                result = json.loads(txt)
                sleep(3)  # to avoid hitting the rate limit

            for article in result['results']:
                if article['abstract_inverted_index'] is not None:
                    article['abstract'] = abstract_inverted_index_to_abstract(article['abstract_inverted_index'])

                del article['abstract_inverted_index']
                article["authorships_display_name"] = ", ".join(
                    [author["author"]["display_name"] for author in article.get("authorships", [])])
                results.append(article)


            print("total articles:",json.dumps(result['meta']["count"]), "currently retrieved:", len(results), "still to retreive:",  result["meta"]["count"] - len(results))
            if 'next_cursor' in result['meta'] and result["meta"]["count"] - len(results) > 0:
                cursor = result['meta']['next_cursor']
        except Exception as e:
            print(f"Error in query_OpenAlex_{api_type}:", e)
            sleep(10)
            break

    return results


def retrieve_articles(queries: list[str], mail: str,api_key,
                           from_publication_date: Optional[str] = None,
                           to_publication_date: Optional[str] = None)->str:
    """
    Retrieves root_set articles from the OpenAlex API based on a list of queries and other optional parameters.

    :param queries: A list of search queries to use for retrieving articles.
    :param mail: The email address to include in the API request for identification.
    :param from_publication_date: Optional start date for filtering results (format: 'YYYY-MM-DD').
    :param to_publication_date: Optional end date for filtering results (format: 'YYYY-MM-DD').

    This function performs the following steps:
    1. Iterates through the list of queries and retrieves articles from the OpenAlex search API .
    2. For each result, it retrieves the incoming and outgoing citations (by using the cite and cited_by APIs).
    3. Converts the abstract from an inverted index format to a readable one, if present.
    4. Stores the results in the `output` dictionary.
    5. Returns the file path of the json containing all the article for that query (the file is stored in the query_results folder).
    Example Usage:
    --------------
    queries = ["quantum computing", "artificial intelligence"]
    mail = "example@example.com"
    from_publication_date = "2020-01-01"
    to_publication_date = "2022-12-31"

    json_result_path = retrieve_articles(queries, mail, from_publication_date, to_publication_date)
    """
    output={}

    for query in queries:

        results=query_OpenAlex("search",query,  mail, api_key, from_publication_date, to_publication_date)
        result_ids = set()
        for result in results:
            result_ids.add(result['id'])

        for result in results:

            result["root_set"] = True
            result['incoming_citations'] =  query_OpenAlex("cite", result['id'], mail, api_key)
            result['outgoing_citations'] = query_OpenAlex("cited_by", result['id'], mail, api_key)

            output[result['id']] = copy.deepcopy(result)

    output_query_path=Path("query_results")
    Path(output_query_path).mkdir(parents=True, exist_ok=True)
    m = hashlib.sha224()
    m.update(str(queries).encode(encoding='UTF-8', errors='strict'))
    cache_key = m.hexdigest()
    output_query_path=Path("query_results")

    json_file_path = output_query_path / f"query_result_{cache_key}.json"
    json.dump(output,open(json_file_path,"w"))

    return str(json_file_path)


def export_articles_to_csv(json_file_path,fields_to_export: Optional[List[str]] = None,export_path=None):
    """
           Exports article information from a JSON file into a CSV file.

           The data to be exported is specified by the user through the `fields_to_export` parameter.
           Articles are categorized as either "root set" or "base set" based on their presence in the
           root_set articles set.

           Parameters:
           - json_file_path (str):
             The path to the JSON file that contains the articles data. The JSON should be structured
             such that it includes article information, including citations (both incoming and outgoing).

           - fields_to_export (Optional[List[str]]):
             A list of strings representing the fields to be included in the CSV export. If not specified,
             all available fields are exported. Available fields are:
               - "id"
               - "doi"
               - "title"
               - "abstract"
               - "publication_date"
               - "language"
               - "type"
               - "type_crossref"
               - "authorships_display_name"
               - "topics"
               - "is_retracted"
               - "cited_by_count"
               - "venue"

             If any fields in `fields_to_export` are not valid, a `ValueError` will be raised.

           - export_path (Optional[str]):
             The path where the CSV file will be saved. If not provided, the default path `"exports"`
             will be used. If a path is provided, the CSV file will be saved in that location.

           Returns:
           - None:
             This function does not return any value. It directly exports the articles data to a CSV file.

           Example usage:
           export_articles_to_csv("path/to/articles.json", fields_to_export=["id", "title", "doi"], export_path="output/articles.csv")
           """

    cache_path = Path("exports")


    field_mapping = lambda article_data: {
        "id": article_data.get("id", "") or "",
        "doi": article_data.get("doi", "") or "",
        "title": article_data.get("title", "") or "",
        "publication_year": article_data.get("publication_year", "") or "",
        "language": article_data.get("language", "") or "",
        "type": article_data.get("type", "") or "",
        "type_crossref": article_data.get("type_crossref", "") or "",
        "authorships display_name": ", ".join(
            [authorship.get("author", {}).get("display_name", "") for authorship in
             article_data.get("authorships", [])
             if isinstance(authorship, dict)]),

        "primary_topic display_name": (article_data.get("primary_topic") or {}).get("display_name", ""),
        "primary_topic subfield display_name": (article_data.get("primary_topic") or {}).get("subfield", {}).get(
            "display_name", ""),
        "primary_topic field display_name": (article_data.get("primary_topic") or {}).get("field", {}).get(
            "display_name", ""),
        "primary_topic domain display_name": (article_data.get("primary_topic") or {}).get("domain", {}).get(
            "display_name", ""),

        "cited_by_count": article_data.get("cited_by_count", "") or "",
        "root_set": article_data.get("root_set", "") or "False",
        "venue": (
                (
                    (article_data.get("primary_location") or {}).get("source")
                ) or {}
        ).get("display_name", ""),
        "publication_date": article_data.get("publication_date", ""),
    }

    available_fields = list(field_mapping({}).keys())

    if fields_to_export:
        invalid_fields = [field for field in fields_to_export if field not in available_fields]
        if invalid_fields:
            raise ValueError(f"The following fields are not available for exporting: {', '.join(invalid_fields)}. Available fields are : {', '.join(available_fields)}")
    else:
        fields_to_export=available_fields

    # Load the articles from the provided JSON file
    json_file_path = Path(json_file_path)  # Convert string to Path object
    if not json_file_path.is_file():
        raise ValueError(f"The input file does not exist: {json_file_path}")

    articles=json.load(open(json_file_path))

    root_set_articles=set()
    for article_id, article_data in articles.items():
        root_set_articles.add(article_id)
    # Determine the CSV file path
    if export_path is None:
        # Use pathlib to handle the export path and ensure cross-platform compatibility
        cache_key = Path(json_file_path).stem.replace(".json", "")
        csv_file_path = cache_path / f"articles_{cache_key}.csv"
    else:
        csv_file_path = Path(export_path)

    # Ensure the cache directory exists
    cache_path.mkdir(parents=True, exist_ok=True)


    with open(csv_file_path, 'w', newline='', encoding='utf-8') as csvfile:
        writer = csv.DictWriter(csvfile, fieldnames=fields_to_export+["scope"])
        writer.writeheader()
        already_save = set()
        for article_id, article_data in tqdm(articles.items(), desc="Exporting articles", unit="article"):

            if article_id not in already_save:
                row = {field: field_mapping(article_data)[field] for field in fields_to_export}
                row["scope"] = "root_set" if article_id in root_set_articles else "base_set"
                writer.writerow(row)
                already_save.add(article_id)

            for citation_type in ['incoming_citations','outgoing_citations']:
                for cite in article_data[citation_type]:
                    if cite["id"] not in already_save:

                        row={field: field_mapping(cite)[field] for field in fields_to_export}
                        row["scope"] = "root_set" if cite["id"] in root_set_articles else "base_set"
                        writer.writerow(row)
                        already_save.add(cite["id"])


    print(f"Export completed. The articles are saved in {csv_file_path}")



def export_articles_to_scopus(json_file_path, include_base_set: bool = True, export_path=None):
    """
           Exports article information from a JSON file into a  Scopus like CSV.



           Parameters:
           - json_file_path (str):
             The path to the JSON file that contains the articles' data. The JSON should be structured
             such that it includes article information, including citations (both incoming and outgoing).

           - include_base_set (Optional[bool]):
             If True, include base set.

           - export_path (Optional[str]):
             The path where the CSV file will be saved. If not provided, the default path `"exports"`
             will be used. If a path is provided, the CSV file will be saved in that location.

           Returns:
           - None:
             This function does not return any value. It directly exports the articles data to a CSV file.

           Example usage:
           export_articles_to_scopus("path/to/articles.json", export_path="output/articles.csv")
           """

    cache_path = Path("exports")


    field_mapping = lambda article_data: {

        "Author": article_data.get("authorships_display_name", ""),
        "Authors full names": "; ".join([
            authorship.get("author", {}).get("display_name", "")+f"({authorship.get('id', '').split('/')[-1]})"
            for authorship in article_data.get("authorships", [])
            if isinstance(authorship, dict)
        ]),
        "Author(s) ID": "; ".join([
            (authorship.get("author", {}).get("id") or "").split("/")[-1]
            for authorship in article_data.get("authorships", [])
            if isinstance(authorship, dict)
        ]),
        "Title": article_data.get("title", ""),
        "Year": article_data.get("publication_year", ""),
        "Source title": (
            ((article_data.get("primary_location") or {}).get("source") or {}).get("display_name", "")
        ),
        "Volume": article_data.get("biblio", {}).get("volume", ""),
        "Issue": article_data.get("biblio", {}).get("issue", ""),
        "Art.No.": "",
        "Page start": article_data.get("biblio", {}).get("first_page", ""),
        "Page end": article_data.get("biblio", {}).get("last_page", ""),
        "Cited by": article_data.get("cited_by_count", ""),
        "DOI": (article_data.get("doi") or "").replace("https://doi.org/", ""),
        "Link": (article_data.get("primary_location") or {}).get("landing_page_url", ""),
        "Affiliations": "; ".join([
            aff
            for authorship in article_data.get("authorships", [])
            for aff in authorship.get("raw_affiliation_strings", [])
        ]),
        "Author(s) with affiliations": "; ".join([
            authorship.get("author", {}).get("display_name", "") + ", " +
            ", ".join(authorship.get("raw_affiliation_strings", []))
            for authorship in article_data.get("authorships", [])
            if isinstance(authorship, dict)
        ]),
        "Abstract": article_data.get("abstract", ""),
        "Author Keywords":"",
        "Index Keywords":"",
        "Molecular Sequence Numbers":"",
        "Chemicals/CAS":"",
        "Tradenames":"",
        "Manufacturers":"",
        "Funding Details": "",  # Non disponibile da OpenAlex
        "Funding Text": "",  # Non disponibile da OpenAlex


        "References": "",
        "Correspondence Address": "",
        "Editors": "",
        "Publisher": "",
        "Sponsors": "",
        "Conference name": "",
        "Conference date": "",
        "Conference location": "",
        "ISSN": ", ".join(
            ((article_data.get("primary_location") or {}).get("source") or {}).get("issn", []) or []
        ),
        "ISBN":"",
        "CODEN":"",
        "PubMed ID":"",
        "Language of Original Document":"",
        "Abbreviated Source Title":"",
        "Document Type": article_data.get("type_crossref", ""),
        "Publication Stage":"",
        "Open Access":"",
        "Source":"scopus",
        "EID": article_data.get("id", "").split("/")[-1],

        "Corresponding Author(s) ID": "; ".join([
            (authorship.get("author", {}).get("id") or "").split("/")[-1]
            for authorship in article_data.get("authorships", [])
            if authorship.get("is_corresponding", False)
        ]),
        "Correspondence Address": "; ".join([
            aff
            for authorship in article_data.get("authorships", [])
            if authorship.get("is_corresponding", False)
            for aff in authorship.get("raw_affiliation_strings", [])
        ]),
        "Publisher": (
            ((article_data.get("primary_location") or {}).get("source") or {}).get("host_organization_name", "")
        ),
        "Language of Original Document": article_data.get("language", ""),
        "Publication Stage": (article_data.get("primary_location") or {}).get("version", ""),
        "Open Access": str(article_data.get("open_access", {}).get("is_oa", False)),
        "OA URL": article_data.get("open_access", {}).get("oa_url", ""),
        "PubMed ID": "",  # Non disponibile da OpenAlex
        "Abbreviated Source Title": "",  # Non disponibile da OpenAlex
        "Source ID": "",  # Non disponibile da OpenAlex
        "Page count": "",  # Non disponibile direttamente
        "Sponsors": "",  # Non disponibile da OpenAlex
        "Publisher ID": "",  # Non disponibile da OpenAlex
        "CODEN": "",  # Non disponibile
        "PubMed Central ID": "",  # Non disponibile
        "Truncated Author(s)": "",  # Non disponibile
    }

    fields_to_export = list(field_mapping({}).keys())



    # Load the articles from the provided JSON file
    json_file_path = Path(json_file_path)  # Convert string to Path object
    if not json_file_path.is_file():
        raise ValueError(f"The input file does not exist: {json_file_path}")

    articles=json.load(open(json_file_path))

    root_set_articles=set()
    for article_id, article_data in articles.items():
        root_set_articles.add(article_id)
    # Determine the CSV file path
    if export_path is None:
        # Use pathlib to handle the export path and ensure cross-platform compatibility
        cache_key = Path(json_file_path).stem.replace(".json", "")
        csv_file_path = cache_path / f"articles_scopus_{cache_key}.csv"
    else:
        csv_file_path = Path(export_path)

    # Ensure the cache directory exists
    cache_path.mkdir(parents=True, exist_ok=True)


    with open(csv_file_path, 'w', newline='', encoding='utf-8') as csvfile:
        writer = csv.DictWriter(csvfile, fieldnames=fields_to_export+["scope"])
        writer.writeheader()
        already_save = set()
        for article_id, article_data in tqdm(articles.items(), desc="Exporting articles", unit="article"):

            if article_id not in already_save:
                row = {field: field_mapping(article_data)[field] for field in fields_to_export}
                row["scope"] = "root_set" if article_id in root_set_articles else "base_set"
                writer.writerow(row)
                already_save.add(article_id)

            for citation_type in ['incoming_citations','outgoing_citations']:
                for cite in article_data[citation_type]:
                    if not include_base_set and cite["id"] not in root_set_articles:
                        continue
                    if cite["id"] not in already_save:
                        row={field: field_mapping(cite)[field] for field in fields_to_export}
                        row["scope"] = "root_set" if cite["id"] in root_set_articles else "base_set"
                        writer.writerow(row)
                        already_save.add(cite["id"])


    print(f"Export completed. The articles are saved in {csv_file_path}")





def export_venues_to_csv(json_file_path,fields_to_export: Optional[List[str]] = None,export_path=None):
    """
           Exports venues information from a JSON file into a CSV file.

           The data to be exported is specified by the user through the `fields_to_export` parameter.
           Venues are aggregate as either "root set" or "base set" based on their presence in the
           root_set articles set.

           Parameters:
           - json_file_path (str):
             The path to the JSON file that contains the venues data. The JSON should be structured
             such that it includes venue information.

           - fields_to_export (Optional[List[str]]):
             A list of strings representing the fields to be included in the CSV export. If not specified,
             all available fields are exported. Available fields are:
               - primary_location source id
               - primary_location source display_name
               - primary_location source issn_l
               - primary_location source issn
               - primary_location source host_organization
               - primary_location source host_organization_name

             If any fields in `fields_to_export` are not valid, a `ValueError` will be raised.

           - export_path (Optional[str]):
             The path where the CSV file will be saved. If not provided, the default path `"exports"`
             will be used. If a path is provided, the CSV file will be saved in that location.

           Returns:
           - None:
             This function does not return any value. It directly exports the articles data to a CSV file.

           Example usage:
           export_venues_to_csv("path/to/articles.json", fields_to_export=["id", "title", "doi"], export_path="output/articles.csv")
           """
    """"""


    author_root_set=Counter()
    author_base_set=Counter()

    cache_path = Path("exports")
    field_mapping = lambda article_data: {
        "primary_location source id": ((article_data.get("primary_location", {}) or {}).get("source",{}) or {}).get("id", ""),
        "primary_location source display_name": ((article_data.get("primary_location", {}) or {}).get("source", {}) or {}).get("display_name", ""),
        "primary_location source issn_l": ((article_data.get("primary_location", {}) or {}).get("source", {}) or {}).get("issn_l", ""),
        "primary_location source issn": ", ".join(
            [ issn for issn in
              ((article_data.get("primary_location", {}) or {}).get("source", {}) or {}).get("issn") or [] ]),
        "primary_location source host_organization": ((article_data.get("primary_location", {}) or {}).get("source", {}) or {}).get("host_organization", ""),
        "primary_location source host_organization_name": ((article_data.get("primary_location", {}) or {}).get("source", {}) or {}).get("host_organization_name", ""),

    }

    available_fields = list(field_mapping({}).keys())

    if fields_to_export:
        invalid_fields = [field for field in fields_to_export if field not in available_fields]
        if invalid_fields:
            raise ValueError(f"The following fields are not available for exporting: {', '.join(invalid_fields)}. Available fields are : {', '.join(available_fields)}")
    else:
        fields_to_export=available_fields

    # Load the articles from the provided JSON file
    json_file_path = Path(json_file_path)  # Convert string to Path object
    if not json_file_path.is_file():
        raise ValueError(f"The input file does not exist: {json_file_path}")

    articles=json.load(open(json_file_path))

    root_set_articles = set(articles.keys())

    # Determine the CSV file path
    if export_path is None:
        # Use pathlib to handle the export path and ensure cross-platform compatibility
        cache_key = Path(json_file_path).stem.replace(".json", "")
        csv_file_path = cache_path / f"venues_{cache_key}.csv"
    else:
        csv_file_path = Path(export_path)

    # Ensure the cache directory exists
    cache_path.mkdir(parents=True, exist_ok=True)

    venues_to_save=[]
    with open(csv_file_path, 'w', newline='', encoding='utf-8') as csvfile:
        writer = csv.DictWriter(csvfile, fieldnames=fields_to_export+["root_set","base_set"])
        writer.writeheader()
        already_done_articles = set()
        for article_id, article_data in tqdm(articles.items(), desc="Exporting articles", unit="article"):
            if article_id not in already_done_articles:

                row = {field: field_mapping(article_data)[field] for field in fields_to_export}
                venue_id=((article_data.get("primary_location", {}) or {}).get("source", {}) or {}).get("id", "")
                if article_id in root_set_articles:
                    author_root_set[venue_id] += 1
                else:
                    author_base_set[venue_id] += 1

                if venue_id not in [venue["primary_location source id"] for venue in venues_to_save]:
                    venues_to_save.append(row)
                already_done_articles.add(article_id)

            for citation_type in ['incoming_citations','outgoing_citations']:
                for cite in article_data[citation_type]:
                    if cite["id"] not in already_done_articles:
                        row={field: field_mapping(cite)[field] for field in fields_to_export}
                        venue_id = ((cite.get("primary_location", {}) or {}).get("source", {}) or {}).get("id","")
                        if cite["id"] in root_set_articles:
                            author_root_set[venue_id] += 1
                        else:
                            author_base_set[venue_id] += 1


                        if venue_id not in [venue["primary_location source id"] for venue in venues_to_save]:
                            venues_to_save.append(row)
                        already_done_articles.add(cite["id"])

        for venue in venues_to_save:
            venue["root_set"]=author_root_set[venue['primary_location source id']]
            venue["base_set"]=author_base_set[venue['primary_location source id']]
            writer.writerow(venue)
    print(f"Export completed. The venues are saved in {csv_file_path}")








def export_authors_to_csv(json_file_path, fields_to_export: Optional[List[str]] = None, export_path=None):
    """
        Exports article information from a JSON file into a CSV file.

        The data to be exported is specified by the user through the `fields_to_export` parameter.
        Articles are categorized as either "root set" or "base set" based on their presence in the
        root_set articles set.

        Parameters:
        - json_file_path (str):
          The path to the JSON file that contains the articles data. The JSON should be structured
          such that it includes article information, including citations (both incoming and outgoing).

        - fields_to_export (Optional[List[str]]):
          A list of strings representing the fields to be included in the CSV export. If not specified,
          all available fields are exported. Available fields are:
            -   "author id"
            -   "author orcid"
            -   "author display_name"
            -   "institutions display_name"
            -   "institutions type"
            -   "countries"


          If any fields in `fields_to_export` are not valid, a `ValueError` will be raised.

        - export_path (Optional[str]):
          The path where the CSV file will be saved. If not provided, the default path `"exports"`
          will be used. If a path is provided, the CSV file will be saved in that location.

        Returns:
        - None:
          This function does not return any value. It directly exports the articles data to a CSV file.

        Example usage:
        export_articles_to_csv("path/to/articles.json", fields_to_export=["id", "orcid", "display_name"], export_path="output/authors.csv")
        """

    author_root_set=Counter()
    author_base_set=Counter()

    field_mapping = lambda data : {
                "author id": (data.get("author", {})  or {}).get("id",""),
                "author orcid": (data.get("author", {}) or {}).get("orcid",""),
                "author display_name": (data.get("author", {}) or {}).get("display_name",""),
                "institutions display_name": " ".join([str(inst["display_name"]) for inst in (data.get("institutions", []) or [])]),
                "institutions type": " ".join([str(inst["type"]) for inst in (data.get("institutions", []) or [])]),
                "countries": " ".join((data.get("countries", []) or []))
            }
    # Available fields for authors
    available_fields = list(field_mapping({}).keys())

    if fields_to_export is not None:
        invalid_fields = [field for field in fields_to_export if field not in available_fields]
        if invalid_fields:
            raise ValueError(
                f"The following fields are not available for exporting: {', '.join(invalid_fields)}. Available fields are : {', '.join(available_fields)}")
    else:
        fields_to_export = available_fields
    # Load the articles from the provided JSON file
    json_file_path = Path(json_file_path)  # Convert string to Path object
    if not json_file_path.is_file():
        raise ValueError(f"The input file does not exist: {json_file_path}")

    articles = json.load(open(json_file_path,"r"))

    # Prepare the output CSV file path
    if export_path is None:
        cache_key = Path(json_file_path).stem.replace(".json", "")
        csv_file_path = Path("exports") / f"authors_{cache_key}.csv"
    else:
        csv_file_path = Path(export_path)

    # Ensure the directory exists
    csv_file_path.parent.mkdir(parents=True, exist_ok=True)

    # Open the CSV file for writing
    already_done_articles=set()
    authors_to_save=[]
    root_set_articles = set(articles.keys())
    with open(csv_file_path, 'w', newline='', encoding='utf-8') as csvfile:
        writer = csv.DictWriter(csvfile, fieldnames=fields_to_export+["root_set","base_set"])
        writer.writeheader()

        already_saved_authors = set()

        # Loop over the articles
        for article_id, article_data in tqdm(articles.items(), desc="Exporting authors", unit="article"):
            if article_id not in already_done_articles:

                for author in article_data.get('authorships', []):
                    author_id = author.get("author").get("id")

                    # Skip authors that have already been saved
                    row = {field: field_mapping(author)[field] for field in fields_to_export}
                    if article_id in root_set_articles:
                        author_root_set[author_id] += 1
                    else:
                        author_base_set[author_id] += 1
                    if author_id not in already_saved_authors:
                        authors_to_save.append(row)

                        already_saved_authors.add(author_id)

            for citation_type in ['incoming_citations','outgoing_citations']:
                for cite in article_data[citation_type]:
                    article_id=cite.get("id")
                    if article_id not in already_done_articles:
                        for author in cite.get('authorships', []):
                            author_id = author.get("author").get("id")
                            # Skip authors that have already been saved
                            row = {field: field_mapping(author)[field] for field in fields_to_export}
                            if article_id in root_set_articles:
                                author_root_set[author_id]+=1
                            else:
                                author_base_set[author_id]+=1
                            if author_id not in already_saved_authors:
                                authors_to_save.append(row)
                                already_saved_authors.add(author_id)
                    already_done_articles.add(author_id)

            already_done_articles.add(article_id)
        for author in authors_to_save:
            author["root_set"]=author_root_set[author['author id']]
            author["base_set"]=author_base_set[author['author id']]
            writer.writerow(author)

    print(f"Export completed. The authors are saved in {csv_file_path}")





def export_institutions_to_csv(json_file_path, fields_to_export: Optional[List[str]] = None, export_path=None):
    """
        Exports institutions information from a JSON file into a CSV file.

        The data to be exported is specified by the user through the `fields_to_export` parameter.
        Articles are categorized as either "root set" or "base set" based on their presence in the
        root articles set.

        Parameters:
        - json_file_path (str):
          The path to the JSON file that contains the institutions' data. The JSON should be structured
          such that it institutions article information

        - fields_to_export (Optional[List[str]]):
          A list of strings representing the fields to be included in the CSV export. If not specified,
          all available fields are exported. Available fields are:
            - institutions id
            - institutions display_name
            - institutions ror
            - institutions country_code,
            - institutions type

          If any fields in `fields_to_export` are not valid, a `ValueError` will be raised.

        - export_path (Optional[str]):
          The path where the CSV file will be saved. If not provided, the default path `"exports"`
          will be used. If a path is provided, the CSV file will be saved in that location.

        Returns:
        - None:
          This function does not return any value. It directly exports the articles data to a CSV file.

        Example usage:
        export_institutions_to_csv("path/to/institutions.json", fields_to_export=["id", "orcid", "display_name"], export_path="output/authors.csv")
        """

    institution_root_set=Counter()
    institution_base_set=Counter()

    field_mapping = lambda data : {
                "institutions id": data.get("id",""),
                "institutions display_name": data.get("display_name",""),
                "institutions ror": data.get("ror",""),
                "institutions country_code": data.get("country_code",""),
                "institutions type": data.get("type",""),
            }
    # Available fields for authors
    available_fields = list(field_mapping({}).keys())

    if fields_to_export is not None:
        invalid_fields = [field for field in fields_to_export if field not in available_fields]
        if invalid_fields:
            raise ValueError(
                f"The following fields are not available for exporting: {', '.join(invalid_fields)}. Available fields are : {', '.join(available_fields)}")
    else:
        fields_to_export = available_fields
    # Load the articles from the provided JSON file
    json_file_path = Path(json_file_path)  # Convert string to Path object
    if not json_file_path.is_file():
        raise ValueError(f"The input file does not exist: {json_file_path}")

    articles = json.load(open(json_file_path,"r"))

    # Prepare the output CSV file path
    if export_path is None:
        cache_key = Path(json_file_path).stem.replace(".json", "")
        csv_file_path = Path("exports") / f"institutions_{cache_key}.csv"
    else:
        csv_file_path = Path(export_path)

    # Ensure the directory exists
    csv_file_path.parent.mkdir(parents=True, exist_ok=True)

    # Open the CSV file for writing
    already_done_institutions=set()
    institutions_to_save=[]
    root_set_articles = set(articles.keys())
    with open(csv_file_path, 'w', newline='', encoding='utf-8') as csvfile:
        writer = csv.DictWriter(csvfile, fieldnames=fields_to_export+["root_set","base_set"])
        writer.writeheader()

        already_saved_institutions = set()

        # Loop over the articles
        for article_id, article_data in tqdm(articles.items(), desc="Exporting authors", unit="article"):
            if article_id not in already_done_institutions:

                for author in article_data.get('authorships', []):
                    author_id = author.get("author").get("id")
                    institutions = author.get('institutions', [])
                    for institution in institutions:
                        institution_id = institution.get("id")

                        # Skip authors that have already been saved
                        row = {field: field_mapping(institution)[field] for field in fields_to_export}
                        if article_id in root_set_articles:
                            institution_root_set[institution_id] += 1
                        else:
                            institution_base_set[institution_id] += 1
                        if institution_id not in already_saved_institutions:
                            institutions_to_save.append(row)

                            already_saved_institutions.add(institution_id)
                        already_done_institutions.add(institution_id)

            for citation_type in ['incoming_citations','outgoing_citations']:
                for cite in article_data[citation_type]:
                    article_id=cite.get("id")
                    if article_id not in already_done_institutions:
                        for author in cite.get('authorships', []):
                            author_id = author.get("author").get("id")
                            institutions = author.get('institutions', [])
                            for institution in institutions:
                                institution_id = institution.get("id")
                                # Skip authors that have already been saved
                                row = {field: field_mapping(institution)[field] for field in fields_to_export}


                                if article_id in root_set_articles:
                                    institution_root_set[institution_id]+=1
                                else:
                                    institution_base_set[institution_id]+=1
                                if institution_id not in already_saved_institutions:
                                    institutions_to_save.append(row)
                                    already_saved_institutions.add(institution_id)
                                already_done_institutions.add(institution_id)

        for institution in institutions_to_save:
            institution["root_set"]=institution_root_set[institution['institutions id']]
            institution["base_set"]=institution_base_set[institution['institutions id']]
            writer.writerow(institution)

    print(f"Export completed. The institutions are saved in {csv_file_path}")



def create_citation_graph(articles: Dict[str, Dict], export_path: Optional[str] = None, base_set=True) -> nx.DiGraph:
    """
    Creates a directed graph from the articles data where nodes represent articles and edges represent citations.

    :param articles: A dictionary of articles where the key is the article ID and the value is the article details.
    :param export_path: Optional path to export the graph in GML format. If None, the graph will not be exported.

    :return: A NetworkX directed graph representing the citation network.

    """
    allowed_keys = ["id","doi","title","display_name","publication_year","primary_topic","authorships","language","type","cited_by_count","root_set"]
    G = nx.DiGraph()
    root_set_articles=set()
    for article_id in articles.keys():
        root_set_articles.add(article_id)

    def save_article_fields_for_gephi(article_data):
        node_data={}
        if "root_set" not in article_data:
            article_data["root_set"]=0
        for key, value in article_data.items():
            if key in ["title","primary_topic","language","type","cited_by_count","root_set"]:
                if value is None:
                    node_data[key] = ""
                elif key == "primary_topic":
                    node_data["primary_topic"]=value.get("display_name","")
                    if "display_name" in value:
                        node_data["primary_subfield"]=value["subfield"].get("display_name","")
                        node_data["primary_field"]=value["field"].get("display_name","")
                        node_data["primary_domain"]=value["domain"].get("display_name","")
                else:
                    node_data[key] = value

        return node_data

    for article_id, article_data in tqdm(articles.items(), desc="Creating citation graph", unit="article"):
        #  Remove unnecessary fields to reduce the graph's size
        _article_data = {key: value for key, value in article_data.items() if key in allowed_keys}

        G.add_node(article_id, **save_article_fields_for_gephi(_article_data),info=json.dumps(_article_data))
        for citing_article in article_data.get('incoming_citations', []):
            citing_article_id = citing_article['id']
            if not base_set and citing_article_id not in root_set_articles:
                continue
            if not G.has_node(citing_article_id):
                _citing_article_article_data = {key: value for key, value in citing_article.items() if key in allowed_keys}
                G.add_node(citing_article_id, **save_article_fields_for_gephi(_citing_article_article_data), info=json.dumps(_citing_article_article_data))
            G.add_edge(citing_article_id, article_id)


        for cited_article in article_data.get('outgoing_citations', []):
            cited_article_id = cited_article['id']
            if not base_set and cited_article_id not in root_set_articles:
                continue
            if not G.has_node(cited_article_id):
                _cited_article_article_data={key: value for key, value in cited_article.items() if key in allowed_keys}
                G.add_node(cited_article_id, **save_article_fields_for_gephi(_cited_article_article_data), info=json.dumps(_cited_article_article_data))
            G.add_edge(article_id, cited_article_id)

    print(f"Performing spring layout")
    pos = nx.spring_layout(G)
    for node, (x, y) in pos.items():
        G.nodes[node]['pos_x'] = str(x)
        G.nodes[node]['pos_y'] = str(y)


    if export_path:
        nx.write_gml(G, export_path)
        print(f"Export completed. The citation graph is saved in {export_path}")

    return G


def create_coauthorship_graph(articles: Dict[str, Dict], export_path: Optional[str] = None, base_set=True) -> nx.Graph:
    """
    Creates an undirected graph representing co-authorship relationships between authors, considering both root set and base set articles.

    :param articles: A dictionary of articles where the key is the article ID and the value is the article details.
    :param export_path: Optional path to export the graph in GML format. If None, the graph will not be exported.
    :param base_set: If True, includes both root_set and base set articles in the co-authorship graph.
    :return: A NetworkX undirected graph representing the co-authorship network.
    """
    G = nx.Graph()
    root_set_articles=set()
    base_set_articles=set()

    for article_id in articles.keys():
        root_set_articles.add(article_id)


    def save_author_fields_for_gephi(article_data):
        node_data={}
        for key, value in article_data.items():
            if key in ["display_name","countries"]:
                if value is None:
                    node_data[key] = ""

                elif key == "countries":
                    node_data["countries"]=", ".join(value)

        return node_data
    # Iterate over each article and collect authors
    for article_id, article_data in tqdm(articles.items(), desc="Creating co-authorship graph", unit="article"):
        authorships = article_data.get("authorships", [])

        # Add authors as nodes to the graph (if not already present)
        authors=[]
        for author in authorships:
            author_id=author["author"]["id"]
            if author_id is not None:
                authors.append(author_id)
                if not G.has_node(author_id):
                    author_info = author
                    G.add_node(author_id, **save_author_fields_for_gephi(author_info),info=json.dumps(author_info))

        # Add edges (co-authorship between each pair of authors)
        for i in range(len(authors)):
            for j in range(i + 1, len(authors)):
                author_1 = authors[i]
                author_2 = authors[j]

                # Add an edge between the two authors (only if it does not already exist)
                if not G.has_edge(author_1, author_2):
                    G.add_edge(author_1, author_2)
                # Determine whether to include the article (root set or base set)
        if base_set:
            # Consider incoming/outgoing citations (base set articles) to connect authors
            for citation_type in ['incoming_citations','outgoing_citations']:
                for citing_article in article_data.get(citation_type, []):
                    citing_article_id = citing_article['id']
                    if citing_article_id in root_set_articles:
                        continue #already done
                    if citing_article_id in base_set_articles:
                        continue #already done

                    base_set_articles.add(citing_article_id)

                    authorships = article_data.get("authorships", [])

                    # Add authors as nodes to the graph (if not already present)
                    authors = []
                    for author in authorships:
                        author_id = author["author"]["id"]
                        if author_id is not None:
                            authors.append(author_id)
                            if not G.has_node(author_id):
                                author_info = author
                                G.add_node(author_id, **save_author_fields_for_gephi(author_info), info=json.dumps(author_info))


                    # Add edges (co-authorship between each pair of authors)
                    for i in range(len(authors)):
                        for j in range(i + 1, len(authors)):
                            author_1 = authors[i]
                            author_2 = authors[j]

                            # Add an edge between the two authors (only if it does not already exist)
                            if not G.has_edge(author_1, author_2):
                                G.add_edge(author_1, author_2)

    print(f"Performing spring layout")
    pos = nx.spring_layout(G)
    for node, (x, y) in pos.items():
        G.nodes[node]['pos_x'] = str(x)
        G.nodes[node]['pos_y'] = str(y)

    # Export the graph in GML format, if a path is provided
    if export_path:
        nx.write_gml(G, export_path)
        print(f"Export completed. The co-authorship graph is saved in {export_path}")


    return G




from itertools import chain
from typing import Dict, List
from datetime import datetime


def check_date_format(dt_str: str) -> bool:
    """
    Checks if a given date string is in the 'YYYY-MM-DD' format.

    This function attempts to parse the input string using the datetime module to ensure it conforms
    to the 'YYYY-MM-DD' format. If the string is correctly formatted, the function returns True.
    Otherwise, it returns False.

    :param dt_str: A string representing a date in the 'YYYY-MM-DD' format.
    :return: True if the string is in the correct 'YYYY-MM-DD' format, otherwise False.

    Example:

    >>> check_date_format('20210714')
    False
    >>> check_date_format('2021-07-14')
    True
    >>> check_date_format('20211314')
    False
    """
    try:
        date = datetime.strptime(dt_str, '%Y-%m-%d')
    except:
        return False
    return True

def abstract_inverted_index_to_abstract(abstract_inverted_index: Dict[str, List[int]]) -> str:
    """
    Generates a string of words sorted by their positional indexes from an inverted index dictionary.

    This function takes an inverted index, where keys are words and values are lists of positions
    where these words occur in the text. It then reconstructs the original text by sorting the words
    based on their positions and concatenating them into a single string.

    :param abstract_inverted_index: A dictionary where keys are words (str) and values are lists of
                                    positions (List[int]) where the words occur in the text.
    :return: A string representing the original text, reconstructed by sorting the words based on
             their positions.

    Example:

    >>> abstract_inverted_index = {
    ... "Despite":[0], "growing":[1], "interest":[2], "in":[3,57,73,110,122],
    ... "Open":[4,201], "Access":[5], "(OA)":[6], "to":[7,54,252], "scholarly":[8,105],
    ... "literature,":[9], "there":[10], "is":[11,107,116,176], "an":[12,34,85,185,199,231],
    ... "unmet":[13], "need":[14,31], "for":[15,42,174,219], "large-scale,":[16],
    ... "up-to-date,":[17], "and":[18,24,77,112,124,144,221,237,256], "reproducible":[19],
    ... "studies":[20], "assessing":[21], "the":[22,104,134,145,170,195,206,213,245],
    ... "prevalence":[23], "characteristics":[25], "of":[26,51,75,83,103,137,141,163,209],
    ... "OA.":[27,168,239], "We":[28,46,97,203,240], "address":[29], "this":[30,114,142],
    ... "using":[32,95,244], "oaDOI,":[33], "open":[35], "online":[36], "service":[37],
    ... "that":[38,89,99,113,147,155], "determines":[39], "OA":[40,56,93,108,138,159,175,210,223,254],
    ... "status":[41], "67":[43], "million":[44], "articles.":[45], "use":[47], "three":[48,58],
    ... "samples,":[49], "each":[50], "100,000":[52], "articles,":[53,152,211], "investigate":[55],
    ... "populations:":[59], "(1)":[60], "all":[61], "journal":[62,70], "articles":[63,71,79,94,164,191,224],
    ... "assigned":[64], "a":[65,250], "Crossref":[66], "DOI,":[67], "(2)":[68], "recent":[69,128],
    ... "indexed":[72], "Web":[74], "Science,":[76], "(3)":[78], "viewed":[80], "by":[81,120,235],
    ... "users":[82,91,157], "Unpaywall,":[84], "open-source":[86], "browser":[87], "extension":[88],
    ... "lets":[90], "find":[92,154], "oaDOI.":[96], "estimate":[98], "at":[100], "least":[101], "28%":[102],
    ... "literature":[106], "(19M":[109], "total)":[111], "proportion":[115], "growing,":[117],
    ... "driven":[118,233], "particularly":[119], "growth":[121], "Gold":[123], "Hybrid.":[125],
    ... "The":[126], "most":[127,171], "year":[129], "analyzed":[130], "(2015)":[131], "also":[132,204],
    ... "has":[133], "highest":[135], "percentage":[136], "(45%).":[139], "Because":[140], "growth,":[143],
    ... "fact":[146], "readers":[148], "disproportionately":[149], "access":[150], "newer":[151],
    ... "we":[153,188], "Unpaywall":[156], "encounter":[158], "quite":[160], "frequently:":[161], "47%":[162],
    ... "they":[165], "view":[166], "are":[167], "Notably,":[169], "common":[172], "mechanism":[173],
    ... "not":[177], "Gold,":[178], "Green,":[179], "or":[180], "Hybrid":[181,238], "OA,":[182], "but":[183],
    ... "rather":[184], "under-discussed":[186], "category":[187], "dub":[189], "Bronze:":[190],
    ... "made":[192], "free-to-read":[193], "on":[194], "publisher":[196], "website,":[197], "without":[198],
    ... "explicit":[200], "license.":[202], "examine":[205], "citation":[207,216], "impact":[208],
    ... "corroborating":[212], "so-called":[214], "open-access":[215], "advantage:":[217], "accounting":[218],
    ... "age":[220], "discipline,":[222], "receive":[225], "18%":[226], "more":[227], "citations":[228],
    ... "than":[229], "average,":[230], "effect":[232], "primarily":[234], "Green":[236], "encourage":[241],
    ... "further":[242], "research":[243], "free":[246], "oaDOI":[247], "service,":[248], "as":[249],
    ... "way":[251], "inform":[253], "policy":[255], "practice.":[257]
    ... }
    >>> abstract = invert_index(abstract_inverted_index)
    >>> print(abstract)
    Despite growing interest in Open Access (OA) to scholarly literature, there is an unmet need for large-scale, up-to-date, and reproducible studies assessing the prevalence and characteristics of OA. We address this using oaDOI, an open online service that determines OA status for 67 million articles. We use three samples, each of 100,000 articles, to investigate three populations: (1) all journal articles assigned a Crossref DOI, (2) recent articles indexed in Web of Science, and (3) articles viewed by users of Unpaywall, an open-source browser extension that lets users find the OA status of articles. We estimate that at least 28% of the scholarly literature (19M articles) is OA. This proportion is growing, driven particularly by growth in Gold and Hybrid OA. The most recent year analyzed (2015) also has the highest percentage (45%). Because of this growth, and the fact that readers disproportionately access newer articles, we find that 47% of the articles they view are OA. Notably, the most common mechanism for OA is not Gold, Green, or Hybrid OA, but rather an under-discussed category we dub Bronze: articles made free-to-read on the publisher website, without an explicit license. We examine the citation impact of OA articles, corroborating the so-called open-access advantage: accounting for article age and discipline, OA articles receive 18% more citations than average, an effect driven primarily by Green OA. We encourage further research using the free oaDOI service, as a way to inform policy and practice.
    """
    words = [[(key, val) for val in val_list] for key, val_list in abstract_inverted_index.items()]
    abstract = list(chain(*words))
    abstract = sorted(abstract, key = lambda x: x[1])
    abstract = [ia[0] for ia in abstract]
    abstract = ' '.join(abstract)

    return abstract




if __name__ == "__main__":
    abstract_inverted_index = {"Despite":[0],"growing":[1],"interest":[2],"in":[3,57,73,110,122],"Open":[4,201],"Access":[5],"(OA)":[6],"to":[7,54,252],"scholarly":[8,105],"literature,":[9],"there":[10],"is":[11,107,116,176],"an":[12,34,85,185,199,231],"unmet":[13],"need":[14,31],"for":[15,42,174,219],"large-scale,":[16],"up-to-date,":[17],"and":[18,24,77,112,124,144,221,237,256],"reproducible":[19],"studies":[20],"assessing":[21],"the":[22,104,134,145,170,195,206,213,245],"prevalence":[23],"characteristics":[25],"of":[26,51,75,83,103,137,141,163,209],"OA.":[27,168,239],"We":[28,46,97,203,240],"address":[29],"this":[30,114,142],"using":[32,95,244],"oaDOI,":[33],"open":[35],"online":[36],"service":[37],"that":[38,89,99,113,147,155],"determines":[39],"OA":[40,56,93,108,138,159,175,210,223,254],"status":[41],"67":[43],"million":[44],"articles.":[45],"use":[47],"three":[48,58],"samples,":[49],"each":[50],"100,000":[52],"articles,":[53,152,211],"investigate":[55],"populations:":[59],"(1)":[60],"all":[61],"journal":[62,70],"articles":[63,71,79,94,164,191,224],"assigned":[64],"a":[65,250],"Crossref":[66],"DOI,":[67],"(2)":[68],"recent":[69,128],"indexed":[72],"Web":[74],"Science,":[76],"(3)":[78],"viewed":[80],"by":[81,120,235],"users":[82,91,157],"Unpaywall,":[84],"open-source":[86],"browser":[87],"extension":[88],"lets":[90],"find":[92,154],"oaDOI.":[96],"estimate":[98],"at":[100],"least":[101],"28%":[102],"literature":[106],"(19M":[109],"total)":[111],"proportion":[115],"growing,":[117],"driven":[118,233],"particularly":[119],"growth":[121],"Gold":[123],"Hybrid.":[125],"The":[126],"most":[127,171],"year":[129],"analyzed":[130],"(2015)":[131],"also":[132,204],"has":[133],"highest":[135],"percentage":[136],"(45%).":[139],"Because":[140],"growth,":[143],"fact":[146],"readers":[148],"disproportionately":[149],"access":[150],"newer":[151],"we":[153,188],"Unpaywall":[156],"encounter":[158],"quite":[160],"frequently:":[161],"47%":[162],"they":[165],"view":[166],"are":[167],"Notably,":[169],"common":[172],"mechanism":[173],"not":[177],"Gold,":[178],"Green,":[179],"or":[180],"Hybrid":[181,238],"OA,":[182],"but":[183],"rather":[184],"under-discussed":[186],"category":[187],"dub":[189],"Bronze:":[190],"made":[192],"free-to-read":[193],"on":[194],"publisher":[196],"website,":[197],"without":[198],"explicit":[200],"license.":[202],"examine":[205],"citation":[207,216],"impact":[208],"corroborating":[212],"so-called":[214],"open-access":[215],"advantage:":[217],"accounting":[218],"age":[220],"discipline,":[222],"receive":[225],"18%":[226],"more":[227],"citations":[228],"than":[229],"average,":[230],"effect":[232],"primarily":[234],"Green":[236],"encourage":[241],"further":[242],"research":[243],"free":[246],"oaDOI":[247],"service,":[248],"as":[249],"way":[251],"inform":[253],"policy":[255],"practice.":[257]}

    abstract= abstract_inverted_index_to_abstract(abstract_inverted_index)
    print(abstract)


import pandas as pd
import matplotlib.pyplot as plt
from matplotlib import colors

import numpy as np
from datetime import datetime
from typing import Dict, Optional
from collections import Counter, defaultdict
from keybert import KeyBERT
from tqdm import tqdm
from matplotlib.lines import Line2D
from matplotlib.colors import to_hex
import spacy

# Carica il modello inglese di spaCy
nlp = spacy.load("en_core_web_sm")

def lemmatize_text(text):
    doc = nlp(text)
    # Ricrea il testo con parole lemmatizzate
    return " ".join([token.lemma_ for token in doc if not token.is_punct])


def _parse_publication_date(date_str: str) -> Optional[datetime]:
    """
    Parses the publication date string and returns a corresponding datetime object.

    Parameters:
    date_str (str): A string representing the publication date in the format "YYYY-MM-DD".

    Returns:
    Optional[datetime]: A datetime object corresponding to the parsed date if successful,
                         or None if the date string is not in the expected format.

    Example:
    >>> _parse_publication_date("2025-01-20")
    datetime.datetime(2025, 1, 20, 0, 0)

    >>> _parse_publication_date("invalid-date")
    None
    """
    try:
        return datetime.strptime(date_str, "%Y-%m-%d")
    except:
        return None


def _aggregate_articles_by_interval(
        data: Dict[str, Dict],
        date_from: Optional[datetime] = None,
        date_to: Optional[datetime] = None,
        interval: str = "month"
) -> pd.DataFrame:
    """
    Aggregates articles by the specified interval (month, quarter, or year).

    This function processes a dictionary of articles and their associated citations,
    aggregates them based on the specified time interval, and returns the count of root set
    and base set articles within each interval.

    Args:
        data (Dict[str, Dict]): A dictionary where the keys are article IDs and the values
                                 are dictionaries containing article details, including
                                 "publication_date", "incoming_citations", and "outgoing_citations".
        date_from (Optional[datetime], optional): The starting date for the aggregation range. Defaults to None.
        date_to (Optional[datetime], optional): The ending date for the aggregation range. Defaults to None.
        interval (str, optional): The time interval for aggregation.
                                  Can be "month", "quarter", or "year". Defaults to "month".

    Returns:
        pd.DataFrame: A DataFrame with the aggregated counts of root set and base set articles for each time interval.

    Raises:
        ValueError: If an unsupported interval is provided.

    Example:
        data = {
            "article1": {
                "publication_date": "2023-01-15",
                "incoming_citations": [{"id": "article2", "publication_date": "2023-02-01"}],
                "outgoing_citations": []
            },
            "article2": {
                "publication_date": "2023-02-01",
                "incoming_citations": [],
                "outgoing_citations": []
            }
        }

        aggregated_data = _aggregate_articles_by_interval(data, date_from=datetime(2023, 1, 1), interval="month")
        print(aggregated_data)
    """
    all_articles = []
    root_set_ids = set(data.keys())

    for article_id, article_data in data.items():

        # Add the root_set article
        pub_date = _parse_publication_date(article_data.get("publication_date"))
        root_set_id=set()
        if pub_date:
            all_articles.append({
                "id": article_id,
                "publication_date": pub_date,
                "is_root_set": True
            })
            root_set_id.add(article_id)

        # Process incoming and outgoing citations
        for citation_key in ["incoming_citations", "outgoing_citations"]:
            for cited_article in article_data.get(citation_key, []):
                cited_id=cited_article["id"]
                if cited_id not in root_set_ids:
                    pub_date = _parse_publication_date(cited_article.get("publication_date"))
                    if pub_date:
                        all_articles.append({
                            "id": cited_id,
                            "publication_date": pub_date,
                            "is_root_set": True if cited_id in root_set_id else False
                        })

    # Create a DataFrame
    df = pd.DataFrame(all_articles).drop_duplicates(subset=["id"])
    # Filter by date range
    if date_from is not None:
        df = df[df["publication_date"] >= date_from]
    if date_to  is not None:
        df = df[df["publication_date"] <= date_to]

    # Add time intervals
    if interval == "month":
        df["interval"] = df["publication_date"].dt.to_period("M")
    elif interval == "quarter":
        df["interval"] = df["publication_date"].dt.to_period("Q")
    elif interval == "year":
        df["interval"] = df["publication_date"].dt.to_period("Y")
    else:
        raise ValueError("Unsupported interval. Use 'month', 'quarter', or 'year'.")

    # Aggregate counts by interval and is_root_set
    aggregated = df.groupby(["interval", "is_root_set"]).size().unstack(fill_value=0)
    aggregated.columns = ["base_set" if not col else "root_set" for col in aggregated.columns]
    for col in ["root_set", "base_set"]:
        if col not in aggregated.columns:
            aggregated[col] = 0

    aggregated = aggregated[["root_set", "base_set"]]

    aggregated = aggregated.reset_index()

    return aggregated


def plot_article_trends(
        articles: dict,
        interval: str,
        color_root_set: str = "#1f78b4",
        color_base_set: str = "#a6cee3",
        date_from: Optional[datetime] = None,
        date_to: Optional[datetime] = None,
        num_ticks: int = 10
) -> None:
    """
    Plots a stacked area chart based on the aggregated article data.

    This function generates a stacked area chart that visualizes the number of root set and
    base set articles over time, aggregated by the specified interval (month, quarter, year).

    Args:
        articles (dict): A dictionary where the keys are article IDs and the values are
                          dictionaries containing article details (including "publication_date",
                          "incoming_citations", and "outgoing_citations").
        interval (str): Time interval for the x-axis labels. Can be "month", "quarter", or "year".
        color_root_set (str, optional): Color for the "root set" articles in the plot. Defaults to "orange".
        color_base_set (str, optional): Color for the "base set" (base) articles in the plot.
                                         Defaults to "skyblue".
        date_from (Optional[pd.Timestamp], optional): Start date for the aggregation range. Defaults to None.
        date_to (Optional[pd.Timestamp], optional): End date for the aggregation range. Defaults to None.
        num_ticks (int, optional): The maximum number of x-axis ticks to display. Defaults to 10.

    Returns:
        None: The function generates and displays the stacked area chart.

    Raises:
        ValueError: If the `interval` is not one of "month", "quarter", or "year".

    Example:
        data = {
            "article1": {
                "publication_date": "2023-01-15",
                "incoming_citations": [{"id": "article2", "publication_date": "2023-02-01"}],
                "outgoing_citations": []
            },
            "article2": {
                "publication_date": "2023-02-01",
                "incoming_citations": [],
                "outgoing_citations": []
            }
        }

        plot_stacked_area_chart(data, interval="month", date_from=pd.Timestamp("2023-01-01"))
    """
    fig, ax = plt.subplots(figsize=(12, 6))
    aggregated_data = _aggregate_articles_by_interval(articles, date_from=date_from, date_to=date_to, interval=interval)

    # Prepare data for plotting
    x = aggregated_data["interval"].astype(str)
    y_root_set = aggregated_data["root_set"]
    y_non_root_set = aggregated_data["base_set"]

    ax.fill_between(x, 0, y_root_set, label="Root Set", color=color_root_set, alpha=0.7)
    ax.fill_between(x, y_root_set, y_root_set + y_non_root_set, label="Base Set", color=color_base_set, alpha=0.7)

    # Configure chart
    ax.set_title(f"Number of articles by {interval}", fontsize=26, fontweight='bold')
    ax.set_xlabel(f"Publication {interval.title()}", fontsize=22)
    ax.set_ylabel("Number of articles", fontsize=22)
    ax.legend(fontsize=20)

    # Adjust x-axis ticks for better readability
    xticks = x[::max(1, len(x) // num_ticks)]  # Reduce the number of ticks to at most num_ticks
    ax.set_xticks(xticks)
    ax.set_xticklabels(xticks, rotation=45, fontsize=20)
    ax.tick_params(axis='y', labelsize=20)
    ax.grid(axis='y', linestyle='--', alpha=0.5)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['left'].set_visible(False)
    plt.tight_layout()
    plt.show()



def _extract_topic(
    data: Dict[str, Dict],
    field_key: Optional[str] = "domain",
) -> pd.DataFrame:
    """
    Extracts topic or concept data from articles for aggregation.

    This function processes the articles dataset, extracting either topics or concepts based
    on the provided key, and returns a DataFrame containing the extracted information.
    Optionally filters by field or domain for topics, or by level for concepts.

    Args:
        data (Dict[str, Dict]): A dictionary where the keys are article IDs and the values are
                                 dictionaries containing article details such as "publication_date",
                                 "topics", "concepts", "incoming_citations", and "outgoing_citations".
        field_key (Optional[str], optional): Must be one of "field" or "domain". Defaults to "domain".

    Returns:
        pd.DataFrame: A DataFrame containing the extracted information, including article ID,
                      publication date, whether the article is root set or base set, and the name of the
                      topic.

    Example:
        data = {
            "article1": {
                "publication_date": "2023-01-15",
                "topics": [{"display_name": "AI", "field": {"display_name": "Computer Science"}}],
                "incoming_citations": [],
                "outgoing_citations": []
            },
            "article2": {
                "publication_date": "2023-02-01",
                "topics": [{"display_name": "Machine Learning", "field": {"display_name": "AI"}}],
                "incoming_citations": [{"id": "article1", "publication_date": "2023-01-10"}],
                "outgoing_citations": []
            }
        }

        extracted_data = _extract_topic_or_concept_data(data, "domain")
        print(extracted_data)
    """
    key="topics"
    extracted_data = []
    root_set_ids = set(data.keys())

    for article_id, article_data in data.items():

        pub_date = _parse_publication_date(article_data.get("publication_date"))
        is_root_set = True

        # Process root set article topics/concepts
        score=0
        field_name = "None"
        for item in article_data.get(key, []):
            if  field_key and field_key in item:
                field_name = item[field_key]["display_name"]
            else:
                if score < item.get("score"):
                    field_name = item.get("display_name")
                    score = item.get("score")
        extracted_data.append({
            "id": article_id,
            "publication_date": pub_date,
            "is_root_set": is_root_set,
            "name": field_name
        })

        # Process citations
        for citation_key in ["incoming_citations", "outgoing_citations"]:
            for cited_article in article_data.get(citation_key, []):
                cited_id=cited_article["id"]
                if cited_id not in root_set_ids:
                    pub_date = _parse_publication_date(cited_article.get("publication_date"))
                    for item in cited_article.get(key, []):
                        if field_key and field_key in item:
                            field_name = item[field_key]["display_name"]
                        else:
                            field_name = item.get("display_name")

                        extracted_data.append({
                            "id": cited_id,
                            "publication_date": pub_date,
                            "is_root_set": False,
                            "name": field_name
                        })

    return pd.DataFrame(extracted_data)

def _aggregate_topics(
    data: pd.DataFrame,
    interval: str = "month",
    date_from: Optional[datetime] = None,
    date_to: Optional[datetime] = None,
    top_n: int = 10
) -> pd.DataFrame:
    """
    Aggregates topics or concepts data by the specified time interval.

    This function filters and aggregates data based on the time interval, and optionally
    filters by date range. It also returns the top N most frequent topics or concepts.

    Args:
        data (pd.DataFrame): DataFrame containing extracted topic or concept data.
        interval (str, optional): Time interval for aggregation ("month", "quarter", "year").
                                  Defaults to "month".
        date_from (Optional[datetime], optional): Start date for filtering the data. Defaults to None.
        date_to (Optional[datetime], optional): End date for filtering the data. Defaults to None.
        top_n (int, optional): The number of top topics or concepts to display. Defaults to 10.

    Returns:
        pd.DataFrame: Aggregated data for plotting, including counts of topics or concepts
                      per time interval, and whether the article is root set or base set.

    Raises:
        ValueError: If the `interval` is not one of "month", "quarter", or "year".

    Example:
        # Assuming `extracted_data` is a DataFrame containing extracted topic or concept data
        aggregated_data = _aggregate_topic_or_concept_data(extracted_data, interval="month", top_n=5)
        print(aggregated_data)
    """
    if date_from:
        data = data[data["publication_date"] >= date_from]
    if date_to:
        data = data[data["publication_date"] <= date_to]

    # Add time intervals
    if interval == "month":
        data["interval"] = data["publication_date"].dt.to_period("M")
    elif interval == "quarter":
        data["interval"] = data["publication_date"].dt.to_period("Q")
    elif interval == "year":
        data["interval"] = data["publication_date"].dt.to_period("Y")
    else:
        raise ValueError("Unsupported interval. Use 'month', 'quarter', or 'year'.")

    # Get top N topics/concepts
    top_items = data["name"].value_counts().head(top_n).index
    data = data[data["name"].isin(top_items)]

    # Aggregate counts by interval, is_root_set, and name
    aggregated = data.groupby(["interval", "is_root_set", "name"]).size().unstack(fill_value=0)
    return aggregated

def plot_topic_trends(
    articles: Dict,
    interval: str = "month",
    top_n_colors: Optional[list] = None,
    show_base_set: bool = True,
    show_root_set: bool = False,
    field_key: Optional[str] = "domain",
    date_from: Optional[datetime] = None,
    date_to: Optional[datetime] = None,
    top_n: int = 10,
    num_ticks: int = 10
) -> None:
    """
    Plots a grouped stacked bar chart for aggregated data based on topics or concepts.

    Args:
        articles (dict): JSON-like dataset containing articles.
        interval (str): Time interval for aggregation ("month", "quarter", "year").
        top_n_colors (Optional[list], optional): List of n_colors to use for the bars.
        show_base_set (bool, optional): Whether to display base set articles.
        show_root_set (bool, optional): Whether to display root set articles.
        field_key (Optional[str], optional): Key to filter by for topics ('field' or 'domain'). Default 'domain'.
        date_from (Optional[datetime], optional): Start date for filtering.
        date_to (Optional[datetime], optional): End date for filtering.
        top_n (int, optional): Number of top topics to display.
        num_ticks (int, optional): Maximum number of ticks on the x-axis.

    Raises:
        ValueError: If invalid parameters are passed, such as unsupported intervals, keys, or field keys.

    Example:
        plot_multiple_stacked_bar_chart(articles, interval="month", top_n=5, show_root_set=True, key="topics")
    """
    fig, ax = plt.subplots(figsize=(14, 8))

    if not  show_base_set and not show_root_set:
        raise ValueError("You must display at least one of the following: root or base set (show_root_set, show_base_set).")
    if interval not in ["month","quarter","year"]:
        raise ValueError("Unsupported interval. Use 'month', 'quarter', or 'year'.")
    if field_key not in ["field", "domain"]:
        raise ValueError("Unsupported 'field_key'.You must use 'field' or 'domain'.")
    key = field_key

    if top_n_colors is None:
        colormap = plt.get_cmap('tab20')
        top_n_colors = [colors.to_hex(c) for c in colormap(np.linspace(0, 1, top_n + 1))]

        # Extract and aggregate data
    data = _extract_topic(articles, key)
    aggregated_data = _aggregate_topics(data, interval=interval, date_from=date_from, date_to=date_to, top_n=top_n)
    if aggregated_data.empty:
        print("No data available for the selected parameters.")
        return
    # Prepare data for plotting
    unique_intervals = aggregated_data.index.get_level_values('interval').unique()
    x = range(len(unique_intervals))  # Numerical indices for x-axis
    interval_labels = [str(interval) for interval in unique_intervals]  # Labels for x-axis

    aggregated_data=aggregated_data.reset_index()
    aggregated_data = aggregated_data.set_index('interval')
    if show_root_set:
        # Ensure all columns and intervals are aligned
        root_set_data = (
            aggregated_data[aggregated_data['is_root_set'] == True]
            .reindex(unique_intervals, fill_value=0)
            .sort_index()
        )
        root_set_data=root_set_data.drop('is_root_set',axis=1)
        if not show_base_set:
            root_set_data = root_set_data.loc[:, (root_set_data != 0).any(axis=0)]

    if show_base_set:

        non_root_set_data = (
            aggregated_data[aggregated_data['is_root_set'] == False]
            .reindex(unique_intervals, fill_value=0)
            .sort_index()
        )
        non_root_set_data=non_root_set_data.drop('is_root_set',axis=1)
        if not show_base_set:
            non_root_set_data = non_root_set_data.loc[:, (non_root_set_data != 0).any(axis=0)]

    # Offsets for grouped bars
    bar_width = 0.4  # Width of each group
    x_root_set = [val - bar_width / 2 for val in x]  # Left bar positions
    x_non_root_set = [val + bar_width / 2 for val in x]  # Right bar positions
    # Plot root_set data
    bottom_root_set = pd.Series(0, index=unique_intervals)

    if show_root_set:

        for column in root_set_data.columns:
            ax.bar(
                x_root_set, root_set_data[column].values,
                label=f"{column}",
                bottom=bottom_root_set.values,
                color=top_n_colors[root_set_data.columns.get_loc(column) % top_n],  # Cycle n_colors
                alpha=0.8, width=bar_width
            )
            bottom_root_set += root_set_data[column]

        if show_base_set and show_root_set:
            # Add symbols on top of bars
            ax.scatter(x_root_set, bottom_root_set.values + root_set_data[column].values, marker=7, color="black", s=30)
    # Plot base_set data
    bottom_non_root_set = pd.Series(0, index=unique_intervals)
    if show_base_set:
        for column in non_root_set_data.columns:
            ax.bar(
                x_non_root_set, non_root_set_data[column].values,
                bottom=bottom_non_root_set.values,
                color=top_n_colors[root_set_data.columns.get_loc(column) % 10],  # Cycle n_colors
                alpha=0.8, width=bar_width
            )
            bottom_non_root_set += non_root_set_data[column]
        if show_base_set and show_root_set:
            # Add symbols on top of bars
            ax.scatter(x_non_root_set, bottom_non_root_set.values + root_set_data[column].values, marker=10, color="black", s=30)

    # Configure chart
    ax.set_title(f"Top {top_n} {key.lower()}s by {interval.lower()}", fontsize=26, fontweight='bold')
    ax.set_xlabel(f"Publication {interval.title()}", fontsize=22)
    ax.set_ylabel("Number of Articles", fontsize=22)

    # Adjust x-axis ticks for better readability
    xticks = list(x)[::max(1, len(x) // num_ticks)]
    interval_labels = list(interval_labels)[::max(1, len(interval_labels) // num_ticks)]
    ax.set_xticks(xticks)
    ax.set_xticklabels(interval_labels, rotation=45,fontsize=20)
    ax.tick_params(axis='y', labelsize=20)
    markers=[]
    markers_labels=[]
    if show_root_set and show_base_set:
        root_set_marker = Line2D([0], [0], marker=7, color='black', markersize=6, linestyle='None', label='Root Set Articles')
        markers.append(root_set_marker)
        markers_labels.append('Root Set Articles')
        base_set_marker = Line2D([0], [0], marker=10, color='black', markersize=6, linestyle='None',
                                  label='Base Set Articles')
        markers.append(base_set_marker)
        markers_labels.append("Base Set Articles")

    # Get existing legend elements
    handles, labels = ax.get_legend_handles_labels()
    by_label = dict(zip(labels, handles))  # Remove duplicate labels

    # Add custom markers to legend
    ax.legend(
        (list(by_label.values()) + markers)[::-1],
        (list(by_label.keys()) + markers_labels)[::-1],
              fontsize=20, loc='upper left', bbox_to_anchor=(1.02, 1))
    ax.grid(axis='y', linestyle='--', alpha=0.5)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['left'].set_visible(False)
    plt.tight_layout()
    plt.show()




def plot_top_authors(
    articles, field_key="domain",
    date_from=None, date_to=None, num_authors=5, by_citations=False,
    show_base_set=True, show_root_set=True, n_colors=None):
    """
    Plots a stacked bar chart showing the top authors and their articles or citations split by topics or concepts.

    Args:
        articles (dict): JSON-like dataset containing articles with metadata.
        field_key (str, optional): Field key to filter topics by ('field' or 'domain').
        date_from (datetime, optional): Start date for filtering articles by publication date.
        date_to (datetime, optional): End date for filtering articles by publication date.
        num_authors (int): Number of top authors to display on the chart.
        by_citations (bool): If True, ranks authors by the number of citations instead of the number of publications.
        show_base_set (bool): If True, includes base set articles in the analysis.
        show_root_set (bool): If True, includes root set articles in the analysis.
        n_colors (list, optional): List of hex color codes for the chart bars. Defaults to a predefined colormap.

    Raises:
        ValueError: If neither `show_base_set` nor `show_root_set` is True.
        ValueError: If `field_key` is invalid.

    Returns:
        None: The function directly displays the chart using matplotlib.

    Example Usage:
        plot_author_stacked_bar_chart(
            articles=my_articles,
            date_from=datetime(2020, 1, 1),
            date_to=datetime(2022, 12, 31),
            num_authors=10,
            by_citations=False,
            n_colors=["#1f77b4", "#ff7f0e", "#2ca02c", "#d62728", "#9467bd"]
        )
    """

    if not  show_base_set and not show_root_set:
        raise ValueError("You must display at least one of the following: root set or base set (show_root_set, show_base_set).")

    if field_key not in ["field", "domain"]:
            raise ValueError("Unsupported 'field_key'. When using key == topics, you must use 'field' or 'domain'.")
    key=field_key
    if n_colors is None:
        colormap = plt.get_cmap('tab10')
        n_colors = [colors.to_hex(c) for c in colormap(np.linspace(0, 1, 10 + 1))]

    author_data = []
    done=set()
    for article_id, article_data in articles.items():
        done.add(article_id)
        pub_date = _parse_publication_date(article_data.get("publication_date"))
        if date_from and (not pub_date or pub_date < date_from):
            continue
        if date_to and (not pub_date or pub_date > date_to):
            continue

        # Find the dominant topic
        field_name = "None"
        if article_data["primary_topic"] is not None:
                field_name = article_data["primary_topic"][field_key]["display_name"]


        # Process authorships
        for authorship in article_data.get("authorships", []):
            author = authorship.get("author", {})
            author_name = author.get("display_name", "Unknown Author")
            if by_citations:
                count = article_data.get("cited_by_count", 0)  # Count citations
            else:
                count = 1  # Count publications
            author_data.append({
                "author": author_name,
                key: field_name,
                "count": count
            })
        if show_base_set:
            for citation_key in ["incoming_citations", "outgoing_citations"]:
                for related_article in article_data.get(citation_key, []):
                    related_article_id = related_article["id"]
                    if related_article_id not in done:
                        done.add(related_article_id)
                        for authorship in related_article.get("authorships", []):
                            author = authorship.get("author", {})
                            author_name = author.get("display_name", "Unknown Author")
                            if by_citations:
                                count = related_article.get("cited_by_count", 0)  # Count citations
                            else:
                                count = 1  # Count publications
                            author_data.append({
                                "author": author_name,
                                key: field_name,
                                "count": count
                            })



    # Create a DataFrame
    df = pd.DataFrame(author_data)
    if df.empty:
        raise Exception("No data using this filter")
    # Aggregate by author and topic
    aggregated = df.groupby(["author", key])["count"].sum().unstack(fill_value=0)

    # Rank authors by total count
    top_authors = aggregated.sum(axis=1).nlargest(num_authors).index
    aggregated = aggregated.loc[top_authors]

    # Plot stacked bar chart
    visible_topics = aggregated.loc[:, (aggregated != 0).any(axis=0)]

    fig, ax = plt.subplots(figsize=(14, 8))
    bottom = pd.Series(0, index=aggregated.index)

    for topic in visible_topics.columns:
        ax.bar(
            aggregated.index, visible_topics[topic],
            bottom=bottom,
            label=topic,
            color=n_colors[visible_topics.columns.get_loc(topic) % len(n_colors)]
        )
        bottom += visible_topics[topic]

    label="number of published articles"
    if by_citations:
        label = "number of citations"

    ax.legend(
        title=f"{key}",
        bbox_to_anchor=(1.05, 1),
        loc='upper left',
        fontsize=22
    )


    # Configure chart
    ax.set_title(f"Top {num_authors} authors by {label}", fontsize=26, fontweight='bold')
    ax.set_xlabel("Authors", fontsize=22)
    ax.set_ylabel("Number of Articles" if not by_citations else "Number of Citations", fontsize=14)
    ax.legend(title=key, bbox_to_anchor=(1.05, 1), loc='upper left', fontsize=16, title_fontsize=22)
    plt.xticks(rotation=45, ha="right",fontsize=20)
    ax.tick_params(axis='y', labelsize=20)
    ax.grid(axis='y', linestyle='--', alpha=0.5)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['left'].set_visible(False)
    plt.tight_layout()
    plt.show()




def plot_top_keywords_from_abstracts(
    articles, date_from=None, date_to=None, show_root_set=True,
    show_base_set=False, top_n=10, ngram_range=(1, 2),color="#abd9e9"
):
    """
    Plots the most frequent keywords extracted from abstracts of articles,
    including citations when `show_base_set=True`.

    Args:
        articles (dict): JSON-like dataset containing articles with metadata.
        date_from (datetime, optional): Start date for filtering articles by publication date.
        date_to (datetime, optional): End date for filtering articles by publication date.
        show_root_set (bool): If True, include root set articles in the analysis.
        show_base_set (bool): If True, include base set articles and their citations.
        top_n (int): Number of top keywords to display.
        ngram_range (tuple): N-gram range for extracting keywords with KeyBERT.
        color (string): hexadecimal color of bars i.e. ,"#abd9e9"

    Raises:
        ValueError: If neither `show_root_set` nor `show_base_set` is True.

    Returns:
        None: The function directly displays the chart using matplotlib.
    """
    if not show_root_set and not show_base_set:
        raise ValueError("You must display at least one of the following: root set or base set (show_root_set, show_base_set).")

    kw_model = KeyBERT()
    keyword_counts = Counter()



    # Process articles
    root_set_article_id=set(articles.keys())

    for article_id, article_data in tqdm(articles.items()):

        pub_date = _parse_publication_date(article_data.get("publication_date"))
        if date_from and (not pub_date or pub_date < date_from):
            continue
        if date_to and (not pub_date or pub_date > date_to):
            continue


        # Extract abstracts for the main article and its base set
        abstracts = []

        # Main article abstract
        main_abstract = article_data.get("abstract", "")
        if main_abstract:
            abstracts.append(main_abstract)

        # Base set: citations and cited articles
        if show_base_set:
            for citation_key in ["incoming_citations", "outgoing_citations"]:
                for related_article in article_data.get(citation_key, []):
                    related_article_id=related_article["id"]
                    pub_date = _parse_publication_date(related_article.get("publication_date"))
                    if date_from and (not pub_date or pub_date < date_from):
                        continue
                    if date_to and (not pub_date or pub_date > date_to):
                        continue
                    if related_article_id in root_set_article_id:
                        continue
                    cited_abstract = related_article.get("abstract", "")
                    if cited_abstract:
                        abstracts.append(cited_abstract)
                    root_set_article_id.add(related_article_id)



        # Extract keywords from abstracts
        for abstract in abstracts:
            lemmatize_abstract = lemmatize_text(abstract)
            keywords = kw_model.extract_keywords(
                lemmatize_abstract, keyphrase_ngram_range=ngram_range
            )
            # Add keywords to the counter (ensuring uniqueness per article)
            keyword_counts.update(set(kw for kw, _ in keywords))

    # Get the top `top_n` keywords
    most_common_keywords = keyword_counts.most_common(top_n)

    # Separate keywords and their counts for plotting
    keywords, counts = zip(*most_common_keywords)

    # Plot the bar chart
    fig, ax = plt.subplots(figsize=(14, 8))
    ax.bar(keywords, counts, color=color, edgecolor='black')

    ax.set_title(f"Top {top_n} Keywords from abstracts", fontsize=26, fontweight='bold')
    ax.set_xlabel("Keywords", fontsize=22)
    ax.set_ylabel("Frequency", fontsize=22)
    plt.xticks(rotation=45, ha="right",fontsize=20)
    ax.tick_params(axis='y', labelsize=20)
    ax.grid(axis='y', linestyle='--', alpha=0.5)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['left'].set_visible(False)
    plt.tight_layout()
    plt.show()


def plot_keyword_trends(
    articles, date_from=None, date_to=None, show_root_set=True,
    show_base_set=False, top_n=5, ngram_range=(1, 2), interval="month", n_colors=None
):
    """
    Plots a line chart showing the trends of top keywords over time.
    """
    if not show_root_set and not show_base_set:
        raise ValueError("You must display at least one of the following: root set or base set (show_root_set, show_base_set).")

    if interval not in ["month", "quarter", "year"]:
        raise ValueError("Invalid interval. Choose from 'month', 'quarter', or 'year'.")

    kw_model = KeyBERT()
    keyword_counts_by_time = defaultdict(Counter)
    root_set_article_id = set(articles.keys())

    for article_id, article_data in tqdm(articles.items()):
        pub_date = _parse_publication_date(article_data.get("publication_date"))

        # Determine time period based on interval
        if interval == "month":
            time_period = pub_date.strftime("%Y-%m")
        elif interval == "quarter":
            quarter = (pub_date.month - 1) // 3 + 1
            time_period = f"{pub_date.year}-Q{quarter}"
        elif interval == "year":
            time_period = pub_date.strftime("%Y")

        # Filtra per date
        if not (date_from and (not pub_date or pub_date < date_from) or date_to and (not pub_date or pub_date > date_to)):
            abstract = article_data.get("abstract", "")
            lemmatize_abstract = lemmatize_text(abstract)
            keywords = kw_model.extract_keywords(lemmatize_abstract, keyphrase_ngram_range=ngram_range)
            keyword_counts_by_time[time_period].update(set(kw for kw, _ in keywords))

        # Base set
        if show_base_set:
            for citation_key in ["incoming_citations", "outgoing_citations"]:
                for related_article in article_data.get(citation_key, []):
                    pub_date = _parse_publication_date(related_article.get("publication_date"))
                    if date_from and (not pub_date or pub_date < date_from):
                        continue
                    if date_to and (not pub_date or pub_date > date_to):
                        continue
                    related_article_id = related_article["id"]
                    if related_article_id in root_set_article_id:
                        continue
                    cited_abstract = related_article.get("abstract", "")
                    if not cited_abstract:
                        continue
                    root_set_article_id.add(related_article_id)
                    if interval == "month":
                        time_period = pub_date.strftime("%Y-%m")
                    elif interval == "quarter":
                        quarter = (pub_date.month - 1) // 3 + 1
                        time_period = f"{pub_date.year}-Q{quarter}"
                    elif interval == "year":
                        time_period = pub_date.strftime("%Y")
                    lemmatize_cited_abstract = lemmatize_text(cited_abstract)
                    keywords = kw_model.extract_keywords(lemmatize_cited_abstract, keyphrase_ngram_range=ngram_range)
                    keyword_counts_by_time[time_period].update(set(kw for kw, _ in keywords))

    # Aggrega dati
    aggregated_counts = defaultdict(lambda: defaultdict(int))
    for time_period, counts in keyword_counts_by_time.items():
        for keyword, count in counts.items():
            aggregated_counts[keyword][time_period] += count

    overall_counts = Counter()
    for keyword, time_data in aggregated_counts.items():
        overall_counts[keyword] += sum(time_data.values())
    top_keywords = [kw for kw, _ in overall_counts.most_common(top_n)]

    df = pd.DataFrame.from_dict(aggregated_counts, orient="index").fillna(0)
    df = df.loc[top_keywords]
    df = df.T.sort_index()

    # --- PLOT ---
    fig, ax = plt.subplots(figsize=(14, 8))

    if n_colors is None:
        colormap = plt.get_cmap('tab10')
        n_colors = [to_hex(c) for c in colormap(np.linspace(0, 1, 10))]

    markers = ["o", "s", "^", "D", "x", "*", "P", "v","^","+"]
    for i, keyword in enumerate(top_keywords):
        y_values = df[keyword]
        color = n_colors[i % len(n_colors)]
        marker = markers[int(i/10) % len(markers)]

        ax.plot(df.index, y_values, marker=marker, linestyle="-", label=keyword, alpha=0.9, color=color)

    ax.grid(axis='y', linestyle='--', alpha=0.5)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['left'].set_visible(False)
    ax.set_title("Keyword Trends Over Time", fontsize=26, fontweight='bold')
    ax.set_xlabel(f"Time Period ({interval})", fontsize=22)
    ax.set_ylabel("Frequency", fontsize=22)
    plt.xticks(rotation=45, ha="right", fontsize=20)
    ax.tick_params(axis='y', labelsize=20)

    handles, labels = ax.get_legend_handles_labels()
    legend_elements = [
        plt.Line2D([0], [0], color=n_colors[i % len(n_colors)], marker = markers[int(i/10) % len(markers)], lw=4, linestyle='-',
                   label=kw) for i, kw in enumerate(top_keywords)
    ]
    ax.legend(handles=legend_elements, title="Keywords", bbox_to_anchor=(1.05, 1),
              loc="upper left", fontsize=20, title_fontsize=22, frameon=False)

    plt.tight_layout()
    plt.show()



import json
from typing import Optional, List, Tuple, Dict, Union
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import matplotlib.patches as mpatches
from matplotlib.lines import Line2D
import pandas as pd
import seaborn as sns
from matplotlib.patches import FancyArrowPatch
import textwrap

def _get_centeroid(arr: List[Tuple[float, float]], position_scale_factor: int =1) -> Tuple[float, float]:
    """
    Calculate the centroid of a list of points.

    :param arr: List of tuples, where each tuple contains (x, y) coordinates of a point.
    :param position_scale_factor: Scaling factor for the position coordinates. Defaults to 1.

    :return: A tuple representing the centroid (x, y).
    """
    length = len(arr)

    sum_x = np.sum([z[0] for z in arr]) * position_scale_factor
    sum_y = np.sum([z[1] for z in arr]) * position_scale_factor
    return sum_x / length, sum_y / length

def _get_limits(arr: List[Tuple[float, float]], limit_border:int = 1) -> Tuple[Tuple[float, float], Tuple[float, float]]:
    """
    Calculate the limits (min and max) for x and y coordinates.

    :param arr: List of tuples, where each tuple contains (x, y) coordinates of a point.
    :param limit_border: Scaling factor for the limits. Defaults to 1.
    :return: A tuple of tuples representing the x and y limits ((x_min, x_max), (y_min, y_max)).
    """
    x_max = np.max([z[0] for z in arr]) * limit_border
    x_min = np.min([z[0] for z in arr]) * limit_border
    y_max = np.max([z[1] for z in arr]) * limit_border
    y_min = np.min([z[1] for z in arr]) * limit_border
    return (x_min, x_max), (y_min, y_max)




def _get_entities_in_clusters_citation_graph(G: nx.DiGraph, topics_level: str = "domain") -> Dict[int, Dict[str, int]]:
    """
    Extracts entities from graph nodes and groups them by their clusters, counting occurrences of each entity.

    Args:
        G (Any): A graph object with nodes containing 'cluster' and 'info' attributes.
                 Each node's 'info' attribute should be a JSON string containing either 'concepts' or 'topics'.
        topics_level (Union[int, None]): The level of topics to filter by ("field" or "domain").

    Returns:
        Dict[int, Dict[str, int]]: A dictionary where each key is a cluster identifier (int) and each value is a
                                   dictionary of entity names (str) with their respective counts (int) within that cluster.
                                   Example:
                                   {
                                       0: {'Social Sciences': 50, 'Medicine': 24, ...},
                                       1: {'Computer Science': 17, 'Engineering': 82, ...},
                                       ...
                                   }
    """
    entries_in_clusters = {}
    for node_id, node_data in G.nodes(data=True):

        if node_data["cluster"] not in entries_in_clusters:
            entries_in_clusters[node_data["cluster"]] = {}


        if topics_level is not None:
            node_info=json.loads(node_data["info"])
            if node_info["primary_topic"] is None:
                continue
            primary_topic=node_info["primary_topic"].get(topics_level,{}).get('display_name',"None")
            if primary_topic not in entries_in_clusters[node_data["cluster"]]:
                entries_in_clusters[node_data["cluster"]][primary_topic] = 0
            entries_in_clusters[node_data["cluster"]][primary_topic] += 1

    return entries_in_clusters


def _get_entities_in_clusters_coauthorship_graph(G: nx.Graph) -> Dict[int, Dict[str, int]]:
    """
    Extracts entities from graph nodes and groups them by their clusters, counting occurrences of each entity.

    Args:
        G (Any): A graph object with nodes containing 'cluster' and 'info' attributes.

    Returns:
        Dict[int, Dict[str, int]]: A dictionary where each key is a cluster identifier (int) and each value is a
                                   dictionary of entity names (str) with their respective counts (int) within that cluster.
                                   Example:
                                   {
                                       0: {'IT': 50, 'FR': 24, ...},
                                       1: {'ES': 17, 'US': 82, ...},
                                       ...
                                   }
    """
    entries_in_clusters = {}
    for node_id, node_data in G.nodes(data=True):

        if node_data["cluster"] not in entries_in_clusters:
            entries_in_clusters[node_data["cluster"]] = {}


        node_info=json.loads(node_data["info"])
        if node_info["countries"] is None or len(node_info["countries"])==0:
            continue

        for country in node_info["countries"]:
            if country not in entries_in_clusters[node_data["cluster"]]:
                entries_in_clusters[node_data["cluster"]][country] = 0
            entries_in_clusters[node_data["cluster"]][country] += 1

    return entries_in_clusters



def _get_top_m_entries(clusters: Dict[int, Dict[str, int]], n: int,top_n_clusters) -> List[str]:
    """
    Aggregates counts of different entries across multiple clusters and returns the top m entries.

    Args:
        clusters (dict): A dictionary where each key is a cluster identifier and each value is a dictionary
                         of entries with their respective counts within that cluster.
                         Example:
                         {
                             0: {'Social Sciences': 507, 'Medicine': 24, ...},
                             1: {'Computer Science': 17, 'Engineering': 82, ...},
                             ...
                         }
        n (int): The number of top entries to return based on their aggregated counts across all clusters.

    Returns:
        list: A list of the top n entries sorted by their aggregated counts in descending order.
              Example: ['Social Sciences', 'Engineering', 'Environmental Science', ...]
    """
    total_counts = {}

    for cluster_id, entries_in_cluster in clusters.items():
        if cluster_id not in top_n_clusters:
            continue
        for entry, count in entries_in_cluster.items():
            if entry not in total_counts:
                total_counts[entry] = 0
            total_counts[entry] += count
    sorted_entries = sorted(total_counts.items(), key=lambda x: x[1], reverse=True)

    return [x[0] for x in sorted_entries[:n]]

def _get_top_n_clusters(G: nx.DiGraph|nx.Graph, n_clusters: int) -> Dict[int, int]:
    """
    Identifies the top N clusters based on the number of nodes in each cluster.

    Args:
        G (Any): A graph-like object containing nodes with cluster information.
        n_clusters (int): The number of top clusters to return.

    Returns:
        Dict[int, int]: A dictionary where keys are cluster identifiers and values are the counts of nodes in each cluster,
                        sorted by the count in descending order.
    """
    top_n_clusters = {}
    for node_id, node_data in G.nodes(data=True):
        if node_data['cluster'] in top_n_clusters:
            top_n_clusters[node_data['cluster']] += 1
        else:
            top_n_clusters[node_data['cluster']] = 1


    top_n_clusters = sorted(top_n_clusters.items(), key=lambda x: x[1], reverse=True)[:n_clusters]

    return dict(top_n_clusters)


def _get_cluster_position(node_sizes, edges_weights, max_node_radius):
    G = nx.DiGraph()

    for node, size in node_sizes.items():
        G.add_node(node, size=size)

    for edge, weight in edges_weights.items():
        source, target = map(int, edge.split())
        G.add_edge(source, target, weight=weight)

    pos = nx.spring_layout(G, k=max_node_radius, iterations=50, weight='weight')
    pos = _prevent_overlap(pos, max_node_radius)

    return _center_graph_positions(pos)


def _prevent_overlap(positions: Dict[int, np.ndarray], max_radius: float) -> Dict[
    int, np.ndarray]:
    """
    Adjusts node positions to prevent overlap by ensuring no nodes are within `2 * max_radius` distance.
    Expands the entire graph outward if any overlap is detected.

    Parameters:
    - positions: A dictionary of node positions {node_id: position_array}.
    - max_radius: The maximum radius of a node, used to set the minimum distance between nodes.

    Returns:
    - adjusted_positions: A dictionary with adjusted positions to avoid overlap.
    """
    adjusted_positions = positions.copy()
    overlap_detected = True

    # Check each pair of nodes for overlap
    while overlap_detected:
        overlap_detected=False
        positions = adjusted_positions.copy()
        for node_a, pos_a in positions.items():
            for node_b, pos_b in positions.items():
                if node_a < node_b:  # Avoid redundant checks
                    dist = np.linalg.norm(pos_a - pos_b)
                    # If nodes are too close, push them apart
                    if dist <= 2.5 * max_radius:
                        overlap_detected = True
                        # If overlap was detected, expand the entire graph outward along each axis
                        if overlap_detected:
                            for node in adjusted_positions:
                                adjusted_positions[node] *= 1.1
                            break

    return adjusted_positions

def _center_graph_positions(cluster_positions: Dict[int, np.ndarray]) -> Dict[int, np.ndarray]:
    """
        Centers the positions of clusters in a graph layout.

        This function calculates the centroid (average position) of all clusters
        and then shifts each cluster's position so that the overall layout is
        centered around the origin.

        Args:
            cluster_positions (Dict[int, np.ndarray]): A dictionary mapping cluster IDs
                to their position vectors (e.g., 2D or 3D coordinates).

        Returns:
            Dict[int, np.ndarray]: A new dictionary with the same keys but with
            position vectors translated to center the entire layout.
    """
    centroid = np.mean(np.array(list(cluster_positions.values())), axis=0)

    centered_positions = {cluster_id: pos - centroid for cluster_id, pos in cluster_positions.items()}

    return centered_positions


def _normalize_degree(data, min_target=1, max_target=10):
    """
    Normalizes degree values to a target range using min-max scaling.

    This function rescales all values in the input dictionary so that they lie
    between `min_target` and `max_target`. If all values are equal, all outputs
    will be set to `max_target`.

    Args:
        data (dict): A dictionary of numerical values to normalize.
        min_target (float, optional): The minimum value in the target range. Defaults to 1.
        max_target (float, optional): The maximum value in the target range. Defaults to 10.

    Returns:
        dict: A new dictionary with the same keys and normalized values.
    """
    min_val = min(data.values())
    max_val = max(data.values())

    if max_val == min_val:
        return {k: max_target for k in data}

    return {k: min_target + (max_target - min_target) * (v - min_val) / (max_val - min_val)
            for k, v in data.items()}

def show_clustered_graph(G,
                         n_clusters: int = 5,
                         m_entries: int = 5, verbose = False,
                         n_cluster_colors: Optional[List[str]] = None,
                         m_entry_colors: Optional[List[str]] = None,
                         min_node_radius: float = 3,
                         max_node_radius: float = 5,
                         min_pie_radius: float = 2,
                         max_pie_radius : float = 3,
                         size_legend_marker=2,
                         size_legend_font=15,
                         size_node_font=13,
                         min_edge_width :int = 1,
                         max_edge_width :int = 200,
                         edge_color: str = '#add8e6',
                         hatch: Optional[str] = '',
                         top_m_entries: Optional[List[str]] = None,
                         bbox_to_anchor_legend_entries: Tuple[float, float, float, float] = (1, 1),
                         bbox_to_anchor_legend_clusters: Tuple[float, float, float, float] = (0, 1),
                         image_size : Tuple[int,int] = (947, 1061),
                         topics_level: Optional[str] = "domain",
                         export_path_png: Optional[str] = None) -> None:
    """
        Visualizes a clustered bibliometric graph with pie charts and legends.

        This function generates a high-level visualization of clusters within a graph `G`,
        where each cluster is represented by a circle, and the composition of entities
        (e.g., fields or domains) is shown using pie charts.

        Parameters:
            G (nx.Graph or nx.DiGraph): The graph representing bibliometric data.
            n_clusters (int): Number of top clusters to display.
            m_entries (int): Number of top entities (e.g., disciplines) to include in pies.
            verbose (bool): If True, prints debug info.
            n_cluster_colors (Optional[List[str]]): Colors for each cluster.
            m_entry_colors (Optional[List[str]]): Colors for each entity in pies.
            min_node_radius (float): Minimum cluster circle size.
            max_node_radius (float): Maximum cluster circle size.
            min_pie_radius (float): Minimum pie chart radius around clusters.
            max_pie_radius (float): Maximum pie chart radius.
            size_legend_marker (int): Marker size in legend for clusters.
            size_legend_font (int): Font size in legend for entries.
            size_node_font (int): Font size for cluster labels.
            min_edge_width (int): Minimum width for inter-cluster edges.
            max_edge_width (int): Maximum width for inter-cluster edges.
            edge_color (str): Color used for edges between clusters.
            hatch (Optional[str]): Pattern to fill the doughnut segments. Default is '', e.g., '/', 'x', or 'o'.
            top_m_entries (Optional[List[str]]): Specific top entries to include in pies.
            bbox_to_anchor_legend_entries (Tuple[float]): Positioning of the entries legend.
            bbox_to_anchor_legend_clusters (Tuple[float]): Positioning of the cluster legend.
            image_size (Tuple[int, int]): Output image size in pixels.
            topics_level (Optional[str]): Taxonomy level ("field" or "domain") for entity clustering.
            export_path_png (Optional[str]): If specified, saves the figure to this path.

        Raises:
            ValueError: If inputs are inconsistent or invalid (e.g., length mismatch of color lists).

        Returns:
            None: Displays or saves the graph visualization.

    """
    if topics_level is not None:
        if topics_level not in ["field","domain"]:
            raise ValueError("topics_level must be either 'field' or 'domain'.")

    if top_m_entries is not None:
        if len(top_m_entries) != m_entries:
            raise ValueError("top entries must be the same size of m_entries.")
    if n_cluster_colors is not None:
        if len(n_cluster_colors) != n_clusters:
            raise ValueError("top n_cluster_colors must be the same size of n_clusters.")
    if m_entry_colors is not None:
        if len(m_entry_colors) != m_entries+1:
            raise ValueError("top m_entry_colors must be the same size of m_entries (one color for other).")

    if top_m_entries != None:
        if len(top_m_entries) != m_entries:
            raise ValueError("top entries must be m_entries.")

    dpi = plt.rcParams['figure.dpi']
    fig_size = (image_size[0]/dpi, image_size[1]/dpi)

    fig = plt.figure(figsize=fig_size)
    # Definisci la griglia di sottografici
    gs = fig.add_gridspec(2, 2, height_ratios=[1, 5])

    # Sottografici per le leggende
    ax_legend_left = fig.add_subplot(gs[0, 0])
    ax_legend_right = fig.add_subplot(gs[0, 1])

    # Sottografico per il grafo
    ax_graph = fig.add_subplot(gs[1, :])

    # Disattivare gli assi per le leggende
    ax_legend_left.axis('off')
    ax_legend_right.axis('off')


    if n_cluster_colors is None:

        n_cluster_colors = n_clusters*["#FFFFFF"]

    if m_entry_colors is None:
        cmap = plt.get_cmap('tab10')
        m_entry_colors = [cmap(i) for i in range(n_clusters+1)] #+1 for Other


    top_n_clusters=_get_top_n_clusters(G, n_clusters)
    top_n_clusters=_normalize_degree(top_n_clusters)


    if G.is_directed():
        entries_in_clusters = _get_entities_in_clusters_citation_graph(G, topics_level)
    else:
        entries_in_clusters = _get_entities_in_clusters_coauthorship_graph(G)
    if verbose:
        print("entries_in_clusters",entries_in_clusters)

    if top_m_entries is None:
        top_m_entries = _get_top_m_entries(entries_in_clusters, m_entries,top_n_clusters)
    map_color_to_entry={}
    for i, entry in enumerate(top_m_entries+["Other"]):
        map_color_to_entry[entry]=m_entry_colors[i]

    cluster_edge_map={}
    for edge in G.edges():
        if G.nodes[edge[0]]['cluster'] not in top_n_clusters.keys() or G.nodes[edge[1]]['cluster'] not in top_n_clusters.keys():
            continue
        if str(G.nodes[edge[0]]['cluster'])+" "+str(G.nodes[edge[1]]['cluster']) not in cluster_edge_map:
            cluster_edge_map[str(G.nodes[edge[0]]['cluster'])+" "+str(G.nodes[edge[1]]['cluster'])]=0
        cluster_edge_map[str(G.nodes[edge[0]]['cluster'])+" "+str(G.nodes[edge[1]]['cluster'])]+=1


    cluster_position_map = _get_cluster_position(top_n_clusters, cluster_edge_map, max_node_radius + max_pie_radius)

    max_radius=0
    ##########################################################################################
    #draw circles

    for i, cluster_id in enumerate(top_n_clusters.keys()):
        radius = min_node_radius +(max_node_radius-min_node_radius) * (top_n_clusters[cluster_id]/max(top_n_clusters.values()))
        x, y = cluster_position_map[cluster_id]
        circle1 = plt.Circle((x, y),
                             radius,
                             ec='white',
                             linewidth=2,
                             color=n_cluster_colors[i],
                             zorder=4,
                             )
        ax_graph.add_patch(circle1)
        ax_graph.annotate(str(int(cluster_id)),
                    fontsize=size_node_font,
                    xy=(x, y),
                    ha='center',
                    va='center',
                    zorder=5,
                    color='black')
        if max_radius<radius:
            max_radius=radius



    ##########################################################################################
    #draw edges
    def draw_curved_edge(ax, x0, y0, x1, y1, weight, max_weight, min_edge_width=0.5, max_edge_width=5, color='gray',
                         clockwise=True):
        # Calcola un offset per la curvatura
        dx = x1 - x0
        dy = y1 - y0
        dist = np.sqrt(dx ** 2 + dy ** 2)

        # Curvature: positiva = anti-orario, negativa = orario
        curvature = 0.2  # puoi modulare in base alla distanza o al peso
        if clockwise:
            curvature = -curvature

        # Usa FancyArrowPatch per arco curvo
        arrow = FancyArrowPatch((x0, y0), (x1, y1),
                                connectionstyle=f"arc3,rad={curvature}",
                                linewidth=min_edge_width + ((max_edge_width - min_edge_width) * (weight / max_weight)),
                                color=color,
                                alpha=0.3,
                                zorder=-100)
        ax.add_patch(arrow)

    max_weight = max(cluster_edge_map.values())
    for community_pair, weight in cluster_edge_map.items():
        source, target = map(int, community_pair.split(" "))
        if source not in cluster_position_map or target not in cluster_position_map:
            continue

        x0, y0 = cluster_position_map[source]
        x1, y1 = cluster_position_map[target]

        # Determina la direzione
        # ad esempio: entrante (target) = orario, uscente (source) = antiorario
        clockwise = True  # True = curvatura in senso orario, False = anti-orario
        # Puoi anche decidere basandoti su qualche proprietà del nodo

        draw_curved_edge(ax_graph, x0, y0, x1, y1, weight, max_weight, min_edge_width=min_edge_width, max_edge_width=max_edge_width,
                         color=edge_color, clockwise=clockwise)





    ###################################################
    ##draw PIES

    handles = {}
    size_legend_markers=[]
    for i, cluster_id in enumerate(cluster_position_map.keys()):
        radius = min_node_radius +(max_node_radius-min_node_radius) * (top_n_clusters[cluster_id]/max(top_n_clusters.values()))

        radius=radius+min_pie_radius + max_pie_radius *( top_n_clusters[cluster_id]/max(top_n_clusters.values()))
        size_legend_markers.append(radius)
        filtered_enties={ "Other" : 0.1}
        for discipline,occurencies in entries_in_clusters[cluster_id].items():
            if discipline in top_m_entries:
                filtered_enties[discipline]=occurencies
            else:
                filtered_enties["Other"]+=occurencies

        x, y = cluster_position_map[cluster_id]

        patches = ax_graph.pie([occurency for occurency in filtered_enties.values()],
                         explode=[0] * len(filtered_enties),
                         startangle=90,
                         center=(x, y),
                         colors=[ map_color_to_entry[entry] for entry in filtered_enties.keys()],
                         counterclock=False,
                         radius=radius)

        if max_radius<radius:
            max_radius=radius
        for i in range(len(patches[0])):
            if list(filtered_enties.values())[i]==0:
                continue
            patches[0][i].set(fill=True,
                              hatch=hatch,
                              edgecolor='#C0C0C0')



    #############################
    ### draw legends


    for entry in top_m_entries+["Other"]:

        handles[entry] = \
            mpatches.Patch(facecolor=map_color_to_entry[entry],
                           label=entry,
                           fill=True,
                           hatch=hatch,
                           edgecolor='#C0C0C0')

    legend_entries = ax_legend_left.legend(handles=[value for x, value in handles.items()],
                                loc='center right',
                                bbox_to_anchor=bbox_to_anchor_legend_entries,
                                fontsize=size_legend_font,
                                frameon=False)



    legend_elements=[]
    for i, cluster_id in enumerate(cluster_position_map.keys()):
        legend_elements.append(Line2D([], [],
                                      markeredgecolor='black',
                                      markerfacecolor=n_cluster_colors[i],
                                      marker='o',
                                      linestyle='None',
                                      markersize=size_legend_markers[i]*size_legend_marker,
                                      label="CLUSTER "+str(cluster_id)))
    legend_clusters=ax_legend_right.legend(handles=legend_elements,
                               loc='center left',
                               bbox_to_anchor=bbox_to_anchor_legend_clusters,
                               fontsize = 16,
                               frameon=False)






    ##############################################
    (x_min, x_max), (y_min, y_max) =_get_limits(cluster_position_map.values())
    plt.axis((x_min,x_max,y_min,y_max))
    plt.xlim((x_min - max_radius*2, x_max +  max_radius*2))
    plt.ylim((y_min - max_radius*2, y_max +  max_radius*2))
    if export_path_png is not None:
        plt.savefig(export_path_png)

    plt.tight_layout()
    plt.show()
    plt.close()


import pandas as pd
import matplotlib.pyplot as plt




def show_cluster_statistics(
        csv_file_path,
        n_cluster_colors,
        n_clusters=5):
    """
    Generate cluster-level statistics and a LaTeX table summarizing the most frequent categories
    (disciplines or countries) for each cluster.

    Parameters
    ----------
    csv_file_path : str
        Path to the CSV file containing the clustering data.
    n_cluster_colors : list of str, optional
        List of HEX color codes to represent clusters. Default is a predefined color palette.
    n_clusters : int, optional
        Number of clusters to display (and number of colors expected).

    Returns
    -------
    stats_df : pandas.DataFrame
        DataFrame containing the computed cluster statistics.
    latex_output : str
        String with the LaTeX table code representing cluster composition.

    Notes
    -----
    The CSV file must contain a 'cluster' column and either:
      - 'primary_topic field display_name' (for citation graph analysis), or
      - 'countries' (for co-authorship graph analysis).
    """

    # --- Check colors ---
    if n_cluster_colors is None:
        n_cluster_colors = n_clusters * ["#FFFFFF"]

    if len(n_cluster_colors) != n_clusters:
        raise ValueError(
            f"The number of colors ({len(n_cluster_colors)}) must match n_clusters ({n_clusters})."
        )

    # --- Read CSV file ---
    df = pd.read_csv(csv_file_path)

    if 'cluster' not in df.columns:
        raise ValueError("The CSV file must contain a 'cluster' column.")
    if ('primary_topic field display_name' not in df.columns) and ('countries' not in df.columns):
        raise ValueError(
            "The CSV file must contain 'primary_topic field display_name' (for citation graphs) "
            "or 'countries' (for co-authorship graphs)."
        )

    # --- Compute statistics per cluster ---
    cluster_stats = []
    for cluster_id, group in df.groupby('cluster'):
        num_entities = len(group)
        if 'primary_topic field display_name' in df.columns:
            type_class = "disciplines"
            entities = "papers"
        else:
            type_class = "countries"
            entities = "authors"

        # Rename to a common 'class' column for easier processing
        group = group.rename(columns={'primary_topic field display_name': 'class', 'countries': 'class'})
        most_common = group['class'].value_counts(normalize=True).head(3)
        entries = list(most_common.index)
        percentages = list((most_common * 100).round(0).astype(int))

        row = {
            'Id': int(cluster_id),
            f'Num of {entities}': num_entities,
        }

        # Fill top 3 most common categories
        for i in range(3):
            if i + 1 == 1:
                post = "st"
            elif i + 1 == 2:
                post = "nd"
            else:
                post = "rd"

            if i < len(entries):
                row[f'{i + 1}{post}'] = f"{entries[i]} ({percentages[i]}\\%)"  # \% for LaTeX
            else:
                row[f'{i + 1}{post}'] = ""

        cluster_stats.append(row)

    stats_df = pd.DataFrame(cluster_stats)

    # Sort clusters by size (descending), then by ID (ascending)
    stats_filtered_df = (
        stats_df
        .sort_values(by=[f'Num of {entities}', 'Id'], ascending=[False, True])
        .head(n_clusters)
        .reset_index(drop=True)
    )

    # --- Generate LaTeX table code ---
    latex_lines = []
    latex_lines.append("\\begin{table}[h!]")
    latex_lines.append("\\centering")
    latex_lines.append("\\tiny")
    latex_lines.append("\\begin{tabular}{lllll}")
    latex_lines.append("\\toprule")
    latex_lines.append(
        f"\\multicolumn{{2}}{{l}}{{\\textbf{{Cluster}}}} & "
        f"\\multicolumn{{3}}{{l}}{{\\textbf{{Most frequent {type_class}}}}}\\\\"
    )
    latex_lines.append("\\cmidrule(lr){1-2} \\cmidrule(lr){3-5}")
    latex_lines.append("Id & Num of papers & 1st & 2nd & 3rd \\\\")
    latex_lines.append("\\midrule")

    # Add table rows with color-coded cluster markers
    for i, row in stats_filtered_df.iterrows():
        color = n_cluster_colors[i]
        latex_color_name = f"color{i}"

        # Convert HEX color to LaTeX RGB format
        r = int(color[1:3], 16) / 255
        g = int(color[3:5], 16) / 255
        b = int(color[5:7], 16) / 255

        latex_lines.append(f"\\definecolor{{{latex_color_name}}}{{rgb}}{{{r:.2f},{g:.2f},{b:.2f}}}")
        latex_lines.append(
            f"\\textcolor{{{latex_color_name}}}{{  \\scalebox{{1.5}}{{$\\bullet$}}  }}~{row['Id']} & "
            f"{row[f'Num of {entities}']} & {row['1st']} & {row['2nd']} & {row['3rd']} \\\\"
        )

    latex_lines.append("\\bottomrule")
    latex_lines.append("\\end{tabular}")
    latex_lines.append(f"\\caption{{Cluster composition and most frequent {type_class}.}}")
    latex_lines.append(f"\\label{{tab:cluster_{entities}}}")
    latex_lines.append("\\end{table}")

    latex_output = "\n".join(latex_lines)

    return stats_df, latex_output



def show_graph_statistics(G,csv_file_path,header_color="#a5b8d7"):
    """
    Displays structural and centrality metric statistics for a given graph using data from a CSV file.

    This function generates the following visualizations:
        - A summary table showing the number of nodes, edges, and graph density.
        - A table with the mean and variance for each available centrality metric.
        - Histograms showing the distribution of each centrality metric.
        - A correlation heatmap between centrality measures.

    Parameters
    ----------
    G : networkx.Graph or networkx.DiGraph
        The graph object whose statistics are to be analyzed.

    csv_file_path : str
        Path to the CSV file containing precomputed metrics for each node.
        The file may include one or more of the following columns:
        'betweenness_centrality', 'closeness_centrality', 'page_rank',
        'in_degree', 'out_degree', 'degree'.

    header_color : str, optional
        Hex color code used for the headers in the generated tables (default: '#a5b8d7').

    Returns
    -------
    None
        This function does not return any value. It displays plots and tables as output.

    """
    df = pd.read_csv(csv_file_path)

    all_available_metrics = ['betweenness_centrality', 'closeness_centrality', 'page_rank', 'in_degree', 'out_degree', 'degree']
    df = df[[col for col in all_available_metrics if col in df.columns]]
    def to_camel_case(text: str) -> str:
        text = text.replace("_", " ")
        words = text.split()
        return ' '.join(word.capitalize() for word in words)

    df.rename(columns=lambda col: to_camel_case(col), inplace=True)

    num_nodes = G.number_of_nodes()
    num_edges = G.number_of_edges()
    density = f"{nx.density(G):.2e}"
    statistics = {}

    for colum in df.columns:
        statistics[colum] = {'mean': df[colum].mean(),
                            'variance': df[colum].var()}
    statistics_table = {
        'Metric':  [f'{col}' for col in statistics],
        'Mean': [f"{stats['mean']:.2e}" for stats in statistics.values()],
        'Variance': [f"{stats['variance']:.2e}" for stats in statistics.values()]
    }


    fig, ax = plt.subplots(figsize=(8, 4))

    ax.axis('tight')
    ax.axis('off')
    y_min=0.9
    vertical_width=0.1
    header = plt.table(cellText=[['Statistic', 'Value']],
                       fontsize=20,
                       cellColours=[[header_color, header_color]],
                       cellLoc='center',
                       bbox=[0, y_min, 1, vertical_width],
                       )
    for cell in header.get_celld().values():
        cell.get_text().set_fontweight('bold')

    for metric,value in zip(['Number of Nodes', 'Number of Edges', 'Density'],[num_nodes,num_edges,density]):
        y_min=y_min-vertical_width
        plt.table(cellText=[[metric, f"{value}"]],
                  cellLoc='center',
                  bbox=[0, y_min, 1, vertical_width]
                  )
    y_min=y_min-vertical_width
    header = plt.table(cellText=[['Statistic', 'Mean', 'Variance']],
                       cellColours=[[header_color, header_color, header_color]],

                       cellLoc='center',
                       bbox=[0,y_min, 1, vertical_width]
                       )
    for cell in header.get_celld().values():
        cell.get_text().set_fontweight('bold')

    for i in range(0,len(statistics)):
        y_min=y_min-vertical_width
        plt.table(cellText=[[statistics_table["Metric"][i],statistics_table["Mean"][i],statistics_table["Variance"][i]]],
                  cellLoc='center',
                  bbox=[0, y_min, 1, vertical_width]
                  )


    plt.show()
    plt.close()

    correlation_matrix = df.corr()

    sns.set_theme(style="white")



    for colum in df.columns:

        plt.figure(figsize=(10, 6))
        ax = sns.histplot(df[colum], bins=30, kde=True, color='b', log_scale=True)
        plt.title(f'Distribution of {colum}', fontsize=20, fontweight='bold')
        plt.xlabel(colum)
        plt.ylabel('Frequency')
        ax.grid(axis='y', linestyle='--', alpha=0.5)
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['bottom'].set_visible(False)
        ax.spines['left'].set_visible(False)
        plt.show()
        plt.close()

    plt.figure(figsize=(8, 6))
    mask = np.tril(np.ones_like(correlation_matrix, dtype=bool))
    np.fill_diagonal(mask, True)

    sns.heatmap(correlation_matrix, mask=mask, annot=True, cmap='coolwarm', center=0, annot_kws={"color": "black"})
    plt.title('Correlation Heatmap of Centrality Measures', fontsize=20, fontweight='bold',pad=30)
    for i, label in enumerate(df.columns):
        wrapped_label = "\n".join(textwrap.wrap(label, width=15))
        if i>0:
            rotation=0
        else:
            rotation=0
        plt.text(0.9 + i, i + 0.5, wrapped_label, ha='right', va='center',  fontsize=13, color="black",rotation=rotation)

    plt.xticks([])
    plt.yticks([])
    plt.xlabel('')
    plt.ylabel('')
    plt.subplots_adjust(left=0.2)
    plt.show()



from pybibx.base import pbx_probe

from pybiblionet.openalex.core import export_articles_to_scopus

json_file_path = "query_results/query_result_1ea020320b230de5a973a39682eaa53dce89a9bb026b441a5f825232.json"
print(json_file_path)  # the articles are saved in a json file inside the query_result folder

export_articles_to_scopus(json_file_path,
                          export_path="scopus/myscopus.csv",
                          include_base_set=True
                          )

filename = "scopus/myscopus.csv"
bibfile = pbx_probe(file_bib=filename,db="scopus",del_duplicated=True)

report= bibfile.eda_bib()


from pybiblionet.bibliometric_analysis.charts import plot_topic_trends, plot_article_trends, \
    plot_top_authors, plot_top_keywords_from_abstracts, plot_keyword_trends
from datetime import datetime
import json
# This script performs various bibliometric analyses on a set of academic articles stored in a JSON file (recovered using test_openalex.py).
# It generates several visualizations:
# 1. Article publication trends over time (by month, quarter, or year).
# 2. Topic trends over time, highlighting root set and base set topics.
# 3. Top authors based on citation count, focusing on the most cited authors.
# 4. Top keywords extracted from article abstracts.
# 5. Keyword trends over time, showing the frequency of keywords and their evolution.
# The script uses `matplotlib` for plotting and supports customization for the time intervals, number of authors,
# and other aspects of the analysis.
if __name__ == "__main__" :


    json_file_path="query_results/query_result_1ea020320b230de5a973a39682eaa53dce89a9bb026b441a5f825232.json"
    articles = json.load(open(json_file_path))



    plot_article_trends(articles,
                        date_from=datetime(2019, 1, 1),
                        interval="month",  # Change to "month" or "quarter" or "year" as needed
                        date_to=datetime(2025, 12, 31),
                        num_ticks=20)
    plot_topic_trends(
        articles,
        top_n_colors= ["#a6cee3","#1f78b4","#b2df8a","#33a02c","#d7191c"],
        show_root_set=True,
        show_base_set=True,
        interval="quarter", # Change to "quarter" or "year"
        date_from=datetime(2019, 1, 1),
        date_to=datetime(2025, 12, 31),
        top_n=10,
        )
    plot_top_authors(
        articles,
        date_from=datetime(2019, 1, 1),
        date_to=datetime(2025, 12, 31),
        num_authors=10,
        by_citations=True,
        show_base_set=False,
        n_colors=["#a6cee3","#1f78b4","#b2df8a","#33a02c","#d7191c","#fdae61"]
    )

    plot_top_keywords_from_abstracts(articles,
                                     show_root_set=True,
                                     show_base_set=True,
                                     date_from=datetime(2019, 1, 1),
                                     date_to=datetime(2025, 12, 31),
                                     )


    plot_keyword_trends(
        articles=articles,
        date_from=datetime(2019, 1, 1),
        date_to=datetime(2021, 1, 1),
        show_root_set=True,
        show_base_set=True,
        top_n=5,
        ngram_range=(1, 2),
        interval="quarter",
    )

















import json

from pybiblionet.openalex.core import string_generator_from_lite_regex, retrieve_articles, create_citation_graph, \
    create_coauthorship_graph, export_articles_to_csv, export_authors_to_csv, export_venues_to_csv, \
    export_institutions_to_csv

# This script retrieves academic articles from OpenAlex related to the concept of "15-minute city"
# using a regex-based query. It exports the retrieved articles and their authors to CSV files,
# then generates and saves two graph representations: a citation graph and a co-authorship graph.
# Finally, it prints the number of nodes and edges in each graph to give an overview of their structure.
if __name__ == "__main__" :


    queries = string_generator_from_lite_regex("(15)( )(minute|min)( )(city)")

    mail = "youremail@example.com"
    api_key = "YOUR_KEY" #how to obtain an api key https://developers.openalex.org/guides/authentication
                         #check your daily usage limit here https://openalex.org/settings/usage
    from_publication_date = "2019-01-01"

    to_publication_date = None
    json_file_path = retrieve_articles(queries, mail, api_key, from_publication_date, to_publication_date)
    print(json_file_path)
    export_institutions_to_csv(json_file_path,
                           fields_to_export=None,
                           export_path="15minute_institutions.csv"
                           )

    export_venues_to_csv(json_file_path,
                           fields_to_export=None,
                           export_path="15minute_venues.csv"
                           )

    export_articles_to_csv(json_file_path,
                           fields_to_export=None,
                           export_path="15minute_articles.csv"
                           )
    export_authors_to_csv(json_file_path,
                           fields_to_export=None,#["author id", "author orcid", "author display_name"],
                           export_path="15minute_authors.csv"
                           )

    articles = json.load(open(json_file_path))
    G_citation=create_citation_graph(articles,"15minute_citation_graph.GML", base_set=True)

    print(G_citation.number_of_nodes(),G_citation.number_of_edges())

    G_coauthorship=create_coauthorship_graph(articles,"15minute_coauthorship_graph.GML", base_set=True)

    print(G_coauthorship.number_of_nodes(),G_coauthorship.number_of_edges())




import pandas as pd

# Carica il CSV
df = pd.read_csv("cluster_and_fields_citation.csv")

# Conta quanti autori per ciascun cluster
cluster_sizes = df['cluster'].value_counts()

# Statistiche di base
num_clusters = cluster_sizes.shape[0]
total_authors = cluster_sizes.sum()
mean_size = cluster_sizes.mean()
median_size = cluster_sizes.median()
min_size = cluster_sizes.min()
max_size = cluster_sizes.max()

# Copertura cumulativa per capire l’80%
sorted_sizes = cluster_sizes.sort_values(ascending=False)
cumulative = sorted_sizes.cumsum() / total_authors
num_clusters_80 = (cumulative <= 0.8).sum() + 1  # +1 per includere quello che supera 80%

# Quota dei 5 cluster più grandi
top5_share = sorted_sizes.head(5).sum() / total_authors * 100

# Stampa report
print("=== Analisi distribuzione cluster ===")
print(f"Numero totale di cluster: {num_clusters}")
print(f"Numero totale di autori: {total_authors}")
print(f"Dimensione media cluster: {mean_size:.2f}")
print(f"Dimensione mediana cluster: {median_size}")
print(f"Minima dimensione cluster: {min_size}")
print(f"Massima dimensione cluster: {max_size}")
print()
print(f"Cluster necessari per coprire l'80% degli autori: {num_clusters_80} ({num_clusters_80/num_clusters*100:.1f}% dei cluster)")
print(f"Quota di autori nei 5 cluster più grandi: {top5_share:.2f}%")
print()
print("=== Alcuni esempi di dimensioni ===")
print("Top 5 cluster:", sorted_sizes.head(5).tolist())
print("Bottom 5 cluster:", sorted_sizes.tail(5).tolist())



import csv
import json
from typing import List

import community.community_louvain as community_louvain
from networkx.algorithms.community.centrality import girvan_newman
import networkx as nx
from infomap import Infomap
from sklearn.cluster import SpectralClustering
import igraph as ig

def _customize_for_network_type(G):
    if G.is_directed():
        type_of_graph = "directed citation"
        available_metrics = {
            "betweenness_centrality": nx.betweenness_centrality,
            "closeness_centrality": nx.closeness_centrality,
            "eigenvector_centrality": nx.eigenvector_centrality,
            "page_rank": nx.pagerank,
            "in_degree": nx.in_degree_centrality,
            "out_degree": nx.out_degree_centrality,
        }
        available_fields = ["id", "doi", "title", "publication_year", "language", "type",
                            "authorships display_name", "primary_topic display_name",
                            "primary_topic subfield display_name",
                            "primary_topic field display_name",
                            "primary_topic domain display_name",
                            "cited_by_count","root_set","type","venue"]

        def extract_fields(data, available_fields):
            field_mapping = {
                "id": data.get("id", "") or "",
                "doi": data.get("doi", "") or "",
                "title": data.get("title", "") or "",
                "publication_year": data.get("publication_year", "") or "",
                "language": data.get("language", "") or "",
                "type": data.get("type", "") or "",
                "authorships display_name": ", ".join(
                    [authorship.get("author", {}).get("display_name", "") for authorship in data.get("authorships", [])
                     if isinstance(authorship, dict)]),

                "primary_topic display_name": (data.get("primary_topic") or {}).get("display_name", ""),
                "primary_topic subfield display_name": (data.get("primary_topic") or {}).get("subfield", {}).get("display_name", ""),
                "primary_topic field display_name": (data.get("primary_topic") or {}).get("field", {}).get("display_name", ""),
                "primary_topic domain display_name": (data.get("primary_topic") or {}).get("domain", {}).get("display_name", ""),

                "cited_by_count": data.get("cited_by_count", "") or "",
                "root_set": data.get("root_set", "") or "False",
                "venue": (data.get("primary_location") or {}).get("display_name", "") or "False",
            }

            extracted_data = {}
            for field in available_fields:
                extracted_data[field] = field_mapping.get(field, None)
            return extracted_data

    else:
        type_of_graph = "undirected co-authorship"
        available_metrics = {
            "betweenness_centrality": nx.betweenness_centrality,
            "closeness_centrality": nx.closeness_centrality,
            "eigenvector_centrality": nx.eigenvector_centrality,
            "page_rank": nx.pagerank,
            "degree": nx.degree_centrality,
        }
        available_fields = ["author id", "author orcid", "author display_name", "institutions display_name", "institutions type", "countries"]

        def extract_fields(data, available_fields):

            field_mapping = {
                "author id": (data.get("author", {})  or {}).get("id",""),
                "author orcid": (data.get("author", {}) or {}).get("orcid",""),
                "author display_name": (data.get("author", {}) or {}).get("display_name",""),
                "institutions display_name": " ".join([inst["display_name"] for inst in (data.get("institutions", []) or [])]),
                "institutions type": " ".join([inst["type"] for inst in (data.get("institutions", []) or [])]),
                "countries": " ".join((data.get("countries", []) or []))
            }

            extracted_data = {}
            for field in available_fields:
                extracted_data[field] = field_mapping.get(field, None)

            return extracted_data

    return type_of_graph,available_metrics,available_fields,extract_fields

def extract_metrics_to_csv(G: nx.DiGraph|nx.Graph, metrics: List[str], fields: List[str], csv_file_path: str):
    """
    Extracts the specified metrics and fields from the graph and saves them to a CSV file.

    :param G: A NetworkX directed or undirected graph.
    :param metrics: A list of metrics to compute (available: 'betweenness_centrality', 'closeness_centrality',
                    'page_rank', 'in_degree', 'out_degree', 'degree').
                    'in_degree', 'out_degree' are available only for directed graphs
                    'degree' is available for undirected graphs

    :param fields: A list of fields to include in the CSV
                   for articles the available fields are: "id", "doi", "title", "publication_year", "language", "type",
                            "authorships display_name", "primary_topic display_name",
                            "primary_topic subfield display_name",
                            "primary_topic field display_name",
                            "primary_topic domain display_name",
                            "cited_by_count","root_set";
                   for authors the available fields are:  "author id", "author orcid", "author display_name", "institutions display_name", "institutions type", "countries")
    :param csv_file_path: The path to the CSV file to save the results.
    """

    type_of_graph, available_metrics, available_fields, extract_fields = _customize_for_network_type(G)

    if fields is None:
        fields = available_fields
    if metrics is None:
        metrics = list(available_metrics.keys())
    if len(fields)==0:
        raise ValueError(f"Choose at least a field from: {', '.join(available_fields)}")
    if len(metrics)==0:
        raise ValueError(f"Choose at least a metric from: {', '.join(available_metrics)}")

    # Check for invalid metrics and fields
    invalid_metrics = [metric for metric in metrics if metric not in available_metrics]
    if invalid_metrics:
        raise ValueError(f"The following metrics are not available in a {type_of_graph} graph: {', '.join(invalid_metrics)}")

    invalid_fields = [field for field in fields if field not in available_fields]

    if invalid_fields:
        raise ValueError(f"The following fields are not available in a {type_of_graph} graph: {', '.join(invalid_fields)}")

    # Compute the metrics
    metric_results = {
        metric: available_metrics[metric](G) if callable(available_metrics[metric]) else dict(available_metrics[metric])
        for metric in metrics}

    # Write to CSV
    with open(csv_file_path, 'w', newline='', encoding='utf-8') as csvfile:
        fieldnames = fields + metrics
        writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
        writer.writeheader()

        for node in G.nodes(data=True):
            node_info=json.loads(node[1]['info'])
            row = extract_fields(node_info,fields)
            for metric in metrics:
                row[metric] = metric_results[metric].get(node[0], '')
            writer.writerow(row)

def extract_clusters_to_csv(G: nx.DiGraph|nx.Graph, fields: List[str], csv_file_path: str):
    """
    Extracts the cluster and selected fields from the graph and saves them to a CSV file.

    :param G: A NetworkX directed or undirected graph.

    :param fields: A list of fields to include in the CSV
                   for articles the available fields are: "id", "doi", "title", "publication_year", "language", "type",
                            "authorships display_name", "primary_topic display_name",
                            "primary_topic subfield display_name",
                            "primary_topic field display_name",
                            "primary_topic domain display_name",
                            "cited_by_count","root_set";
                   for authors the available fields are:  "author id", "author orcid", "author display_name", "institutions display_name", "institutions type", "countries")
    :param csv_file_path: The path to the CSV file to save the results.
    """



    type_of_graph, _, available_fields, extract_fields = _customize_for_network_type(G)
    if fields is None:
        fields = available_fields
    if len(fields)==0:
        raise ValueError(f"Choose at least a field from: {', '.join(available_fields)}")

    invalid_fields = [field for field in fields if field not in available_fields]


    if invalid_fields:
        raise ValueError(f"The following fields are not available in a {type_of_graph} graph: {', '.join(invalid_fields)}")

    # Write to CSV
    with open(csv_file_path, 'w', newline='', encoding='utf-8') as csvfile:
        fieldnames = fields + ["cluster"]
        writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
        writer.writeheader()

        for node in G.nodes(data=True):
            node_info = json.loads(node[1]['info'])


            row = extract_fields(node_info, fields)

            row["cluster"] = node[1]["cluster"]
            writer.writerow(row)

def cluster_graph(G: nx.DiGraph|nx.Graph, algorithm: str = 'louvain', num_clusters=5) -> nx.DiGraph|nx.Graph:
    """
    Clusters the nodes in the graph using the specified clustering algorithm and returns a new graph with cluster numbers as node attributes.

    :param G: A NetworkX directed or undirected graph.
    :param algorithm: The clustering algorithm to use ('louvain', 'girvan_newman','infomap', 'spectral_clustering', 'stochastic_block_model').
    :param num_cluster: The number of clusters to create when Spectral Clustering is used. Default is 5.
    :return: A NetworkX graph with cluster numbers as a new node attribute.


    """
    if algorithm == 'louvain':
        partition = community_louvain.best_partition(G.to_undirected(),random_state=87)
        for node, cluster in partition.items():
            G.nodes[node]['cluster'] = cluster
    elif algorithm == 'girvan_newman':
        comp = girvan_newman(G)
        clusters = tuple(sorted(c) for c in next(comp))
        for cluster_num, cluster_nodes in enumerate(clusters):
            for node in cluster_nodes:
                G.nodes[node]['cluster'] = cluster_num
    elif algorithm == 'infomap':

        mapping_realid2fakeid = {}
        mapping_fakeid2realid = {}

        for i, node in enumerate(G.nodes()):
            mapping_realid2fakeid[node] = i
            mapping_fakeid2realid[i] = node

        H = nx.relabel_nodes(G, mapping_realid2fakeid, copy=True)
        im = Infomap(silent=False)
        im.add_networkx_graph(H)
        im.run()

        nodes = im.get_nodes(depth_level=1)
        for node in nodes:
            G.nodes[mapping_fakeid2realid[node.node_id]]['cluster'] = node.module_id
    elif algorithm == 'spectral_clustering':
        adj_matrix = nx.to_numpy_array(G)
        spectral = SpectralClustering(n_clusters=min(num_clusters, len(G)), affinity='precomputed', random_state=87)
        labels = spectral.fit_predict(adj_matrix)

        for node, cluster in zip(G.nodes(), labels):
            G.nodes[node]['cluster'] = cluster

    elif algorithm == 'stochastic_block_model':
        if G.is_directed():
            mode="directed"
        else:
            mode="undirected"

        adjacency_matrix = nx.to_numpy_array(G)
        g = ig.Graph.Weighted_Adjacency(adjacency_matrix.tolist(), mode=mode)
        sbm = g.community_infomap()  # SBM with Infomap
        for cluster_num, cluster_nodes in enumerate(sbm):
            for node in cluster_nodes:
                G.nodes[list(G.nodes())[node]]['cluster'] = cluster_num
    else:
        raise ValueError("Unsupported algorithm. Choose from 'louvain' or 'girvan_newman'.")


    return G




import networkx as nx
from pybiblionet.bibliometric_analysis.core import extract_metrics_to_csv, cluster_graph, extract_clusters_to_csv
from pybiblionet.bibliometric_analysis.charts_network import show_clustered_graph, show_cluster_statistics, \
    show_graph_statistics

# This script performs several operations on a citation graph of academic articles (recovered using test_openalex.py):
# 1. Loads a graph from a GML file using `networkx`.
# 2. Extracts a set of centrality and degree metrics from the graph (betweenness centrality, closeness centrality,
#    eigenvector centrality, page rank, in-degree, and out-degree), and saves them along with additional article fields
#    into a CSV file.
# 3. Applies the Louvain clustering algorithm to the graph to detect communities (clusters) of articles.
# 4. Saves the extracted cluster information (e.g., node IDs) to another CSV file.
# 5. Visualizes the clustered graph using `show_clustered_graph`, showing different clusters of articles.
# 6. Generates and displays a bar chart with statistics about the clusters, such as the number of nodes per cluster.
# This analysis allows for exploring the structure of the network of academic citations and understanding the distribution
# of various metrics and clusters within the data.

if __name__ == "__main__":
    # Read the graph

    json_file_path="query_results/query_result_1ea020320b230de5a973a39682eaa53dce89a9bb026b441a5f825232.json"

    for network_type in ["citation","coauthorship",]:

        if network_type=="citation":
            metrics = ['betweenness_centrality', 'closeness_centrality',
                    'page_rank', 'in_degree', 'out_degree']
            fields  = ["id", "doi", "title", "publication_year", "language", "type",
            "authorships display_name", "primary_topic display_name",
            "primary_topic subfield display_name",
            "primary_topic field display_name",
            "primary_topic domain display_name",
            "cited_by_count", "root_set"]
        else:
            metrics = ['betweenness_centrality', 'closeness_centrality',
                    'page_rank', 'degree']
            fields  = None



        network_file_name= f"15minute_{network_type}_graph.GML"

        G = nx.read_gml(network_file_name)
        print("Graph loaded.")
        print(G.number_of_nodes(), G.number_of_edges())


        csv_file_path = f"metrics_and_fields_{network_type}.csv"

        print("Extracting metrics...")
        extract_metrics_to_csv(G, metrics, fields, csv_file_path)
        show_graph_statistics(G,csv_file_path)

        print("Metrics extracted and saved to CSV.")

        print("\nClustering the graph...")

        clustered_graph = cluster_graph(G, algorithm='louvain',)


        # you can save the clustered graph so it can be imported into visualization or network analysis tools (e.g., Gephi)
        # nx.write_gml(clustered_graph, "clustered_"+network_file_name)
        # clustered_graph=nx.read_gml("clustered_"+network_file_name)

        csv_file_path = "cluster_and_fields_citation.csv"
        extract_clusters_to_csv(clustered_graph, fields, csv_file_path)

        print("Clusters extracted and saved to CSV.")
        # Visualize clusters
        print("Visualizing clusters...")
        n_cluster_colors=["#d7191c","#fdae61","#ffffbf","#abd9e9","#2c7bb6"]
        m_entry_colors = ["#8c510a","#d8b365","#f6e8c3","#c7eae5","#5ab4ac","#01665e"]
        stats_df, latex_output=show_cluster_statistics(csv_file_path, n_cluster_colors=n_cluster_colors, n_clusters=5)



        print(latex_output)
        show_clustered_graph(clustered_graph, image_size=(800, 800),
                             n_clusters=5,
                             topics_level="field",
                             m_entry_colors=m_entry_colors,
                             n_cluster_colors=n_cluster_colors,

                             )

